#!/usr/bin/env python3

"""
Created on 15 Apr 2020

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

source repo: scs_analysis

DESCRIPTION
The sample_collator utility is used to separate input JSON document dependent values according to the upper and lower
bounds of an independent value. For each column, assignment follows the rule:

lower bound <= value < upper bound

The upper and lower bounds for the data set should be specified, along with a delta size. The number of columns required
to service this domain is calculated automatically. The path identifying the leaf node in the input document for both
the independent and dependent fields must be specified.

For each input document, if either the independent or dependent variable values cannot be interpreted ans a number,
then the document is ignored. If either of the fields is not present, the sample_collator utility terminates.

Two collators are provided in this package: sample_collator collates into separate columns (collate to columns),
whereas csv_collator collates into separate CSV files (collate to rows).

SYNOPSIS
sample_collator.py -x IND_PATH [-n NAME] -y DEP_PATH [-l LOWER_BOUND] -u UPPER_BOUND -d DELTA [-v]

EXAMPLES
csv_reader.py -v -l 1 scs-pb1-3-ref-opc-r1-error-2019-09-27T11-03-41+01-00-15min-exegesis-error.csv | \
sample_collator.py -v -x "ref.pmx.PM25 Processed Measurement (µg/m³)" -y error.pm2p5 -u 60 -d 10

DOCUMENT EXAMPLE - INPUT
{"rec": "2019-10-11T10:15:00Z",
"opc": {"pmx": {"tag": "scs-pb1-3", "src": "R1", "val": {"per": 9.9, "pm1": 2.2, "pm2p5": 8.4, "pm10": 12.0}}},
"error": {"pm1": 1.517, "pm2p5": 3.561, "pm10": 3.429}}

DOCUMENT EXAMPLE - OUTPUT
{"rec": "2019-10-11T10:15:00Z",
"opc": {"pmx": {"tag": "scs-pb1-3", "src": "R1", "val": {"per": 9.9, "pm1": 2.2, "pm2p5": 8.4, "pm10": 12.0}}},
"error": {"pm1": 1.517, "pm2p5": {"src": 3.561, "PM25_0-10": 3.561, "PM25_10-20": null, "PM25_20-30": null,
"PM25_30-40": null, "PM25_40-50": null, "PM25_50-60": null, "PM25_60-70": null}, "pm10": 3.429}}

SEE ALSO
scs_analysis/csv_collator
"""

import sys

from scs_analysis.cmd.cmd_sample_collator import CmdSampleCollator

from scs_core.data.json import JSONify
from scs_core.data.path_dict import PathDict
from scs_core.data.sample_delta import SampleDelta


# --------------------------------------------------------------------------------------------------------------------

if __name__ == '__main__':

    document_count = 0
    processed_count = 0

    # ----------------------------------------------------------------------------------------------------------------
    # cmd...

    cmd = CmdSampleCollator()

    if not cmd.is_valid():
        cmd.print_help(sys.stderr)
        exit(2)

    if cmd.verbose:
        print("sample_collator: %s" % cmd, file=sys.stderr)

    try:
        # ------------------------------------------------------------------------------------------------------------
        # resources...

        deltas = SampleDelta.deltas(cmd.lower, cmd.upper, cmd.delta, name=cmd.name)


        # ------------------------------------------------------------------------------------------------------------
        # run...

        # collect data...
        for line in sys.stdin:
            datum = PathDict.construct_from_jstr(line)
            paths = datum.paths()

            document_count += 1

            if cmd.ind_path not in paths:
                print("sample_collator: independent path not in datum: %s" % cmd.ind_path, file=sys.stderr)
                exit(1)

            try:
                ind_value = float(datum.node(cmd.ind_path))
            except ValueError:
                continue                                            # independent value is NaN - skip this row

            if cmd.dep_path not in paths:
                print("sample_collator: dependent path not in datum: %s" % cmd.dep_path, file=sys.stderr)
                exit(1)

            try:
                dep_value = float(datum.node(cmd.dep_path))
            except ValueError:
                continue                                            # dependent value is NaN - skip this row

            target = PathDict()

            for path in paths:
                if path != cmd.dep_path:
                    target.copy(datum, path)
                    continue

                target.append('.'.join((cmd.dep_path, 'src')), dep_value)

                for delta in deltas:
                    value = dep_value if delta.includes(ind_value) else None
                    target.append('.'.join((cmd.dep_path, delta.description())), value)

            print(JSONify.dumps(target))
            sys.stdout.flush()

            processed_count += 1


    # ----------------------------------------------------------------------------------------------------------------
    # end...

    except KeyError as ex:
        print("sample_collator: KeyError: %s" % ex, file=sys.stderr)

    except KeyboardInterrupt:
        print(file=sys.stderr)

    finally:
        if cmd.verbose:
            print("sample_collator: documents: %d processed: %d" % (document_count, processed_count), file=sys.stderr)



#!/usr/bin/env python3

"""
Created on 29 Jun 2021

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

DESCRIPTION
The alert_status utility is used to report on the alerts generated by a specific alert specification. Alert causes are:

* L - Below lower threshold
* U - Above upper threshold
* N - Null value

SYNOPSIS
alert_status.py { -l | -d [-c CAUSE] } [-i INDENT] [-v] ID

EXAMPLES
alert_status.py -d -cL 123

DOCUMENT EXAMPLE
{"id": 77, "rec": "2021-09-07T11:40:00Z", "cause": null, "val": 589.6}

SEE ALSO
scs_analysis/alert
scs_analysis/monitor_auth
"""

import requests
import sys

from scs_analysis.cmd.cmd_alert_status import CmdAlertStatus

from scs_core.aws.client.monitor_auth import MonitorAuth
from scs_core.aws.manager.alert_status_finder import AlertStatusFinder

from scs_core.data.json import JSONify
from scs_core.data.datetime import LocalizedDatetime

from scs_core.sys.http_exception import HTTPException
from scs_core.sys.logging import Logging

from scs_host.sys.host import Host


# --------------------------------------------------------------------------------------------------------------------

if __name__ == '__main__':

    logger = None
    auth = None
    response = None
    report = None

    try:
        # ------------------------------------------------------------------------------------------------------------
        # cmd...

        cmd = CmdAlertStatus()

        if not cmd.is_valid():
            cmd.print_help(sys.stderr)
            exit(2)

        Logging.config('alert_status', verbose=cmd.verbose)
        logger = Logging.getLogger()

        logger.info(cmd)


        # ------------------------------------------------------------------------------------------------------------
        # resources...

        if not MonitorAuth.exists(Host):
            logger.error('MonitorAuth not available.')
            exit(1)

        try:
            auth = MonitorAuth.load(Host, encryption_key=MonitorAuth.password_from_user())
        except (KeyError, ValueError):
            logger.error('incorrect password.')
            exit(1)

        finder = AlertStatusFinder(requests, auth)
        logger.info(finder)


        # ------------------------------------------------------------------------------------------------------------
        # run...

        if cmd.latest:
            response = finder.find(cmd.id, None, cmd.response_mode())
            report = response.alert_statuses[0] if response.alert_statuses else None

        if cmd.history:
            response = finder.find(cmd.id, cmd.cause, cmd.response_mode())
            report = sorted(response.alert_statuses)

        # report...
        if report is not None:
            print(JSONify.dumps(report, indent=cmd.indent))

        if cmd.history:
            logger.info('retrieved: %s' % len(response.alert_statuses))


    except KeyboardInterrupt:
        print(file=sys.stderr)

    except HTTPException as ex:
        now = LocalizedDatetime.now().utc().as_iso8601()
        logger.error("%s: HTTP response: %s (%s) %s" % (now, ex.status, ex.reason, ex.data))
        exit(1)



#!/usr/bin/env python3

"""
Created on 10 Feb 2021

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

source repo: scs_analysis

DESCRIPTION
The sample_distance utility is used to find the distance (in kilometres, to the nearest metre) between a given position,
and the position in each of the input GPS JSON documents. A command flag specifies the path to the node within the
document that is to be examined.

The quality of the GPS fix may be taken into account: if a quality is specified, then any GPS fix with a quality
(rounded to the nearest integer) below that level is reported as a null distance.

A simple, spherical model of the earth is used.

SYNOPSIS
sample_distance.py -p LAT LNG [-i ISO] [-q QUALITY] [-v] GPS_PATH

EXAMPLES
csv_reader.py -v scs-ph1-10-status-H2-15min.csv | \
sample_distance.py -v -p 51.4889752 -0.4418752 -q 1 val.gps | \
csv_writer.py -v scs-ph1-10-distance-H2-15min.csv

DOCUMENT EXAMPLE - INPUT
{"rec": "2020-12-09T21:00:00Z", "val": {"gps": {"pos": [51.48877673, -0.44155907], "elv": 33.3, "qual": 1.0},
"tag": "scs-ph1-10"}

DOCUMENT EXAMPLE - OUTPUT
{"rec": "2020-12-09T21:00:00Z", "gps": {"pos": [51.48877673, -0.44155907], "elv": 33.3, "qual": 1}, "dist": 0.031}

RESOURCES
Getting distance between two points based on latitude/longitude
https://stackoverflow.com/questions/19412462/getting-distance-between-two-points-based-on-latitude-longitude
"""

import sys

from scs_analysis.cmd.cmd_sample_distance import CmdSampleDistance

from scs_core.data.json import JSONify
from scs_core.data.path_dict import PathDict

from scs_core.position.gps_datum import GPSDatum
from scs_core.position.position import Position

from scs_core.sys.logging import Logging


# --------------------------------------------------------------------------------------------------------------------

if __name__ == '__main__':

    document_count = 0
    processed_count = 0

    # ----------------------------------------------------------------------------------------------------------------
    # cmd...

    cmd = CmdSampleDistance()

    if not cmd.is_valid():
        cmd.print_help(sys.stderr)
        exit(2)

    Logging.config('sample_distance', verbose=cmd.verbose)
    logger = Logging.getLogger()

    logger.info(cmd)

    try:
        # ------------------------------------------------------------------------------------------------------------
        # resources...

        origin = Position(cmd.lat, cmd.lng)


        # ------------------------------------------------------------------------------------------------------------
        # run...

        min_datum = None

        for line in sys.stdin:
            datum = PathDict.construct_from_jstr(line)

            if datum is None:
                continue

            document_count += 1

            if not datum.has_sub_path(cmd.iso):
                logger.error("ISO node '%s' not present: %s" % (cmd.iso, line.strip()))
                exit(1)

            if not datum.has_sub_path(cmd.path):
                logger.error("GPS node '%s' not present: %s" % (cmd.path, line.strip()))
                exit(1)

            gps_node = datum.node(cmd.path)

            try:
                gps = GPSDatum.construct_from_jdict(gps_node)
                distance = gps.distance(origin, minimum_acceptable_quality=cmd.quality)

            except TypeError:
                gps = None
                distance = None

            report = PathDict()
            report.append('rec', datum.node(cmd.iso))
            report.append('gps', gps)
            report.append('dist', distance)

            print(JSONify.dumps(report))

            processed_count += 1


    # ----------------------------------------------------------------------------------------------------------------
    # end...

    except KeyboardInterrupt:
        print(file=sys.stderr)

    finally:
        logger.info("documents: %d processed: %d" % (document_count, processed_count))



#!/usr/bin/env python3

"""
Created on 31 Jul 2021

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

DESCRIPTION
The baseline_conf utility is used to specify the operating parameters of the baseline utility.

Aggregation periods must be whole divisors of one hour, i.e. 1, 2, 3, 4, 5, 6, 10, 12, 15, 20, or 30 minutes.

Note that if the start time is a later hour than the end time, then the start time is interpreted as that hour
on the previous day.

Any number of separate baseline_conf files may be stored. The --list flag lists those that are available.

SYNOPSIS
baseline_conf.py [-a] { -z | -l | -c NAME [-t TIMEZONE_NAME] [-s START] [-e END] [-p AGGREGATION]
[-g GAS MINIMUM] [-r GAS] } [-i INDENT] [-v]

EXAMPLES
./baseline_conf.py -c freshfield -t Europe/London -s 23 -e 8 -a 5 -g NO2 10

DOCUMENT EXAMPLE
{"timezone": "Europe/London", "start-hour": 17, "end-hour": 8, "aggregation-period": {"interval": 5, "units": "M"},
"gas-minimums": {"CO": 200, "CO2": 420, "H2S": 5, "NO": 10, "NO2": 10, "SO2": 5}}

FILES
~/SCS/conf/NAME_baseline_conf.json

SEE ALSO
scs_analysis/baseline
scs_dev/gases_sampler
"""

import sys

from scs_analysis.cmd.cmd_baseline_conf import CmdBaselineConf

from scs_core.aws.client.access_key import AccessKey
from scs_core.aws.client.client import Client
from scs_core.aws.manager.s3_manager import S3PersistenceManager

from scs_core.data.json import JSONify
from scs_core.data.recurring_period import RecurringMinutes

from scs_core.estate.baseline_conf import BaselineConf

from scs_core.location.timezone import Timezone

from scs_core.sys.logging import Logging

from scs_host.sys.host import Host


# --------------------------------------------------------------------------------------------------------------------

if __name__ == '__main__':

    conf = None
    key = None

    # ----------------------------------------------------------------------------------------------------------------
    # cmd...

    cmd = CmdBaselineConf()

    if not cmd.is_valid():
        cmd.print_help(sys.stderr)
        exit(2)

    Logging.config('baseline_conf', verbose=cmd.verbose)
    logger = Logging.getLogger()

    logger.info(cmd)


    # ----------------------------------------------------------------------------------------------------------------
    # resources...

    # PersistenceManager...
    if cmd.aws:
        if not AccessKey.exists(Host):
            logger.error('AccessKey not available.')
            exit(1)

        try:
            key = AccessKey.load(Host, encryption_key=AccessKey.password_from_user())
        except (KeyError, ValueError):
            logger.error('incorrect password.')
            exit(1)

        client = Client.construct('s3', key)
        resource_client = Client.resource('s3', key)

        persistence_manager = S3PersistenceManager(client, resource_client)

    else:
        persistence_manager = Host

    # GasModelConf...
    if not cmd.duplicate():
        conf = BaselineConf.load(persistence_manager, name=cmd.conf_name)


    # ----------------------------------------------------------------------------------------------------------------
    # run...

    if cmd.zones:
        for zone in Timezone.zones():
            print(zone, file=sys.stderr)
        exit(0)

    if cmd.list:
        for conf_name in BaselineConf.list(persistence_manager):
            print(conf_name, file=sys.stderr)
        exit(0)

    if cmd.duplicate():
        source = BaselineConf.load(persistence_manager, name=cmd.duplicate_from)

        if source is None:
            logger.error("no configuration found for '%s'" % cmd.duplicate_from)
            exit(2)

        conf = source.duplicate(cmd.duplicate_to)
        conf.save(persistence_manager)

    if cmd.set():
        if conf is None and not cmd.is_complete():
            logger.error("no configuration is stored - you must therefore set all fields.")
            exit(2)

        if cmd.timezone and not Timezone.is_valid(cmd.timezone):
            logger.error("unrecognised timezone name: %s." % cmd.timezone)
            exit(2)

        if cmd.aggregation_period and cmd.aggregation_period not in RecurringMinutes.valid_intervals():
            logger.error("aggregation period must be one of: %s." % str(RecurringMinutes.valid_intervals()))
            exit(2)

        timezone = conf.timezone if cmd.timezone is None else cmd.timezone
        start_hour = conf.start_hour if cmd.start_hour is None else cmd.start_hour
        end_hour = conf.end_hour if cmd.end_hour is None else cmd.end_hour
        aggregation_period = conf.aggregation_period if cmd.aggregation_period is None else \
            RecurringMinutes(cmd.aggregation_period)

        if start_hour == end_hour:
            logger.error("the start and end hours may not be the same.")
            exit(2)

        minimums = {} if conf is None else conf.minimums

        conf = BaselineConf(cmd.conf_name, timezone, start_hour, end_hour, aggregation_period, minimums)

        if cmd.set_gas_name:
            if cmd.set_gas_name not in BaselineConf.supported_gases():
                logger.error("the gas '%s' is not supported." % cmd.set_gas_name)
                exit(2)

            conf.set_minimum(cmd.set_gas_name, cmd.set_gas_minimum)

        if cmd.remove_gas:
            conf.remove_minimum(cmd.remove_gas)

        conf.save(persistence_manager)

    if conf:
        print(JSONify.dumps(conf, indent=cmd.indent))



"""
Created on 11 Jul 2016

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

source repo: scs_analysis
"""

import optparse


# --------------------------------------------------------------------------------------------------------------------

class CmdSampleDistance(object):
    """unix command line handler"""

    def __init__(self):
        """
        Constructor
        """
        self.__parser = optparse.OptionParser(usage="%prog -p LAT LNG [-i ISO] [-q QUALITY] [-v] GPS_PATH",
                                              version="%prog 1.0")

        # compulsory...
        self.__parser.add_option("--position", "-p", type="float", nargs=2, action="store", dest="position",
                                 help="position (in degrees)")

        # optional...
        self.__parser.add_option("--iso-path", "-i", type="string", nargs=1, action="store", default="rec", dest="iso",
                                 help="path for ISO 8601 datetime output (default 'rec')")

        self.__parser.add_option("--quality", "-q", type="int", nargs=1, action="store", dest="quality",
                                 help="minimum acceptable GPS quality")

        self.__parser.add_option("--verbose", "-v", action="store_true", dest="verbose", default=False,
                                 help="report narrative to stderr")

        self.__opts, self.__args = self.__parser.parse_args()


    # ----------------------------------------------------------------------------------------------------------------

    def is_valid(self):
        if self.position is None or self.path is None:
            return False

        return True


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def position(self):
        return self.__opts.position


    @property
    def lat(self):
        return None if self.__opts.position is None else self.__opts.position[0]


    @property
    def lng(self):
        return None if self.__opts.position is None else self.__opts.position[1]


    @property
    def iso(self):
        return self.__opts.iso


    @property
    def quality(self):
        return self.__opts.quality


    @property
    def verbose(self):
        return self.__opts.verbose


    @property
    def path(self):
        return self.__args[0] if len(self.__args) > 0 else None


    # ----------------------------------------------------------------------------------------------------------------

    def print_help(self, file):
        self.__parser.print_help(file)


    def __str__(self, *args, **kwargs):
        return "CmdSampleDistance:{position:%s, iso:%s, quality:%s, verbose:%s, path:%s}" % \
               (self.position, self.iso, self.quality, self.verbose, self.path)



"""
Created on 20 Apr 2021

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)
"""

import optparse

from scs_core.aws.manager.configuration_check_finder import ConfigurationCheckRequest

from scs_core.estate.configuration_check import ConfigurationCheck


# --------------------------------------------------------------------------------------------------------------------

class CmdConfigurationMonitorCheck(object):
    """unix command line handler"""

    def __init__(self):
        """
        Constructor
        """
        codes = ' | '.join(ConfigurationCheck.result_codes())

        self.__parser = optparse.OptionParser(usage="%prog { -c TAG | [-t TAG [-e]] [-r RESULT] [-o] } [-i INDENT] "
                                                    "[-v]", version="%prog 1.0")

        # operations...
        self.__parser.add_option("--check", "-c", type="string", action="store", dest="check",
                                 help="check the device with TAG now")

        # filters...
        self.__parser.add_option("--tag-filter", "-t", type="string", action="store", dest="tag_filter",
                                 help="the (partial) tag of the device(s)")

        self.__parser.add_option("--exactly", "-e", action="store_true", dest="exact_match", default=False,
                                 help="exact match for tag")

        self.__parser.add_option("--result", "-r", type="string", nargs=1, action="store", dest="result_code",
                                 help="match the result { %s }" % codes)

        # output...
        self.__parser.add_option("--tags-only", "-o", action="store_true", dest="tags_only", default=False,
                                 help="report device tags only")

        self.__parser.add_option("--indent", "-i", type="int", nargs=1, action="store", dest="indent",
                                 help="pretty-print the output with INDENT")

        self.__parser.add_option("--verbose", "-v", action="store_true", dest="verbose", default=False,
                                 help="report narrative to stderr")

        self.__opts, self.__args = self.__parser.parse_args()


    # ----------------------------------------------------------------------------------------------------------------

    def is_valid(self):
        if self.check_tag and (self.tag_filter or self.result_code or self.tags_only):
            return False

        if self.exact_match and self.tag_filter is None:
            return False

        if self.result_code and self.result_code not in ConfigurationCheck.result_codes():
            return False

        return True


    def result(self):
        return ConfigurationCheck.result_string(self.result_code)


    def response_mode(self):
        return ConfigurationCheckRequest.MODE.TAGS_ONLY if self.tags_only else ConfigurationCheckRequest.MODE.FULL


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def check_tag(self):
        return self.__opts.check


    @property
    def tag_filter(self):
        return self.__opts.tag_filter


    @property
    def exact_match(self):
        return self.__opts.exact_match


    @property
    def result_code(self):
        return self.__opts.result_code


    @property
    def tags_only(self):
        return self.__opts.tags_only


    @property
    def indent(self):
        return self.__opts.indent


    @property
    def verbose(self):
        return self.__opts.verbose


    # ----------------------------------------------------------------------------------------------------------------

    def print_help(self, file):
        self.__parser.print_help(file)


    def __str__(self, *args, **kwargs):
        return "CmdConfigurationMonitorCheck:{check:%s, tag_filter:%s, exact_match:%s, result_code:%s, tags_only:%s, " \
               "indent:%s, verbose:%s}" % \
               (self.check_tag, self.tag_filter, self.exact_match, self.result_code, self.tags_only,
                self.indent, self.verbose)



#!/usr/bin/env python3

"""
Created on 13 Oct 2016

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

source repo: scs_analysis

DESCRIPTION
The multi_chart utility is used to display a Matplotlib categorical chart for one or more data sources. The
data sources share a common y-axis scale. Data is provided by a sequence of JSON documents on stdin. Each charting
source is specified by a path to a leaf node in the JSON document.

An optional "batch" ("-b") flag can be set, causing the plotting only to take place when all data points have been
received.

Note that the chart is a simple approximation to a timeline chart - values are plotted successively, with no account
taken of the interval between samples.

Depending on operating system, it may be necessary to use a matplotlibrc file, which specifies the Matplotlib
back-end graphics system.

SYNOPSIS
multi_chart.py [-b] [-x POINTS] [-y MIN MAX] [-e] [-t TITLE] [-v] PATH_1 .. PATH_N

EXAMPLES
socket_receiver.py | multi_chart.py val.opc_n2.pm10 val.opc_n2.pm2p5 val.opc_n2.pm1 -x 120 -e

FILES
~/SCS/scs_analysis/src/scs_analysis/matplotlibrc

SEE ALSO
scs_analysis/histo_chart
scs_analysis/single_chart

BUGS
On some operating systems, the chart will remain as the uppermost window until all data have been received.
"""

import sys
import warnings

from scs_analysis.chart.multi_chart import MultiChart
from scs_analysis.cmd.cmd_multi_chart import CmdMultiChart

from scs_core.data.json import JSONify
from scs_core.data.path_dict import PathDict

from scs_core.sync.line_reader import LineReader

from scs_core.sys.logging import Logging


# --------------------------------------------------------------------------------------------------------------------

if __name__ == '__main__':
    warnings.filterwarnings("ignore", module="matplotlib")

    chart = None
    proc = None

    # ----------------------------------------------------------------------------------------------------------------
    # cmd...

    cmd = CmdMultiChart()

    if not cmd.is_valid():
        cmd.print_help(sys.stderr)
        exit(2)

    Logging.config('multi_chart', verbose=cmd.verbose)
    logger = Logging.getLogger()

    logger.info(cmd)

    try:
        # ------------------------------------------------------------------------------------------------------------
        # resources...

        # reader...
        reader = LineReader(sys.stdin.fileno())
        logger.info(reader)

        # chart...
        chart = MultiChart(cmd.title, cmd.batch_mode, cmd.x, cmd.y[0], cmd.y[1], *cmd.paths)
        logger.info(chart)
        logger.info("backend: %s" % chart.backend())

        logger.error("terminate this utility by closing the chart window.")


        # ------------------------------------------------------------------------------------------------------------
        # run...

        proc = reader.start()

        for line in reader.lines:
            if chart.closed:
                break

            if line is None:
                chart.pause()
                continue

            datum = PathDict.construct_from_jstr(line)

            if datum is None:
                break

            if cmd.echo:
                print(JSONify.dumps(datum))
                sys.stdout.flush()

            chart.plot(datum)


    # ----------------------------------------------------------------------------------------------------------------
    # end...

        if not chart.closed:
            chart.render()

    except KeyboardInterrupt:
        print(file=sys.stderr)


    # ----------------------------------------------------------------------------------------------------------------
    # close...

    finally:
        if proc:
            proc.terminate()

        if chart is not None and not chart.closed:
            logger.info("closing.")
            chart.close(None)



#!/usr/bin/env python3

"""
Created on 24 Nov 2021

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

source repo: scs_analysis

DESCRIPTION
The cognito_identity utility is used to create, update and retrieve the AWS Cognito identity for the user.

If the --Create function is used, an email is sent to the new user. The verification link in the email must be
excercised in order for the account to gain a CONFIRMED status.

SYNOPSIS
cognito_identity.py [-c CREDENTIALS] | -C | -R | -U } [-i INDENT] [-v]

EXAMPLES
./cognito_identity.py -R

DOCUMENT EXAMPLE
{"username": "8", "creation-date": "2021-11-24T12:51:12Z", "confirmation-status": "CONFIRMED", "enabled": true,
"email": "bruno.beloff@southcoastscience.com", "given-name": "Bruno", "family-name": "Beloff", "is-super": true}

SEE ALSO
scs_analysis/cognito_credentials

RESOURCES
https://docs.aws.amazon.com/cognito/latest/developerguide/signing-up-users-in-your-app.html
https://docs.aws.amazon.com/cognito/latest/developerguide/user-pool-settings-policies.html
"""

import requests
import sys

from scs_analysis.cmd.cmd_cognito_identity import CmdCognitoIdentity

from scs_core.aws.security.cognito_account_manager import CognitoCreateManager, CognitoUpdateManager

from scs_core.aws.security.cognito_finder import CognitoFinder
from scs_core.aws.security.cognito_login_manager import CognitoLoginManager
from scs_core.aws.security.cognito_user import CognitoUserCredentials, CognitoUserIdentity

from scs_core.data.datum import Datum
from scs_core.data.json import JSONify

from scs_core.sys.http_exception import HTTPException, HTTPConflictException
from scs_core.sys.logging import Logging

from scs_host.comms.stdio import StdIO
from scs_host.sys.host import Host


# --------------------------------------------------------------------------------------------------------------------

if __name__ == '__main__':

    logger = None
    gatekeeper = None
    credentials = None
    auth = None
    finder = None
    report = None

    try:
        # ------------------------------------------------------------------------------------------------------------
        # cmd...

        cmd = CmdCognitoIdentity()

        if not cmd.is_valid():
            cmd.print_help(sys.stderr)
            exit(2)

        Logging.config('cognito_identity', verbose=cmd.verbose)
        logger = Logging.getLogger()

        logger.info(cmd)


        # ------------------------------------------------------------------------------------------------------------
        # auth...

        if not cmd.create:
            gatekeeper = CognitoLoginManager(requests)

            # CognitoUserCredentials...
            if not CognitoUserCredentials.exists(Host, name=cmd.credentials_name):
                logger.error("Cognito credentials not available.")
                exit(1)

            try:
                password = CognitoUserCredentials.password_from_user()
                credentials = CognitoUserCredentials.load(Host, name=cmd.credentials_name, encryption_key=password)
            except (KeyError, ValueError):
                logger.error("incorrect password")
                exit(1)

            try:
                auth = gatekeeper.login(credentials)

            except HTTPException as ex:
                logger.error(ex.data)
                exit(1)


        # ------------------------------------------------------------------------------------------------------------
        # resources...

        if not cmd.create:
            finder = CognitoFinder(requests)


        # ------------------------------------------------------------------------------------------------------------
        # run...

        if cmd.retrieve:
            report = finder.get_self(auth.id_token)

        if cmd.create:
            # create...
            given_name = StdIO.prompt("Enter given name: ")
            family_name = StdIO.prompt("Enter family name: ")
            email = StdIO.prompt("Enter email address: ")
            password = StdIO.prompt("Enter password: ")

            if not given_name or not given_name:
                logger.error("Given name and family name are required.")
                exit(1)

            if not Datum.is_email_address(email):
                logger.error("The email address '%s' is not valid." % email)
                exit(1)

            if not CognitoUserIdentity.is_valid_password(password):
                logger.error("The password '%s' is not valid." % password)
                exit(1)

            report = CognitoUserIdentity(None, None, None, None, email, given_name, family_name, password)

            manager = CognitoCreateManager(requests)
            report = manager.create(report)

            # create credentials...
            credentials = CognitoUserCredentials(cmd.credentials_name, report.email, report.password)
            credentials.save(Host)

        if cmd.update:
            # find...
            identity = finder.get_self(auth.id_token)

            # update identity...
            given_name = StdIO.prompt("Enter given name (%s): ", default=identity.given_name)
            family_name = StdIO.prompt("Enter family name (%s): ", default=identity.family_name)
            email = StdIO.prompt("Enter email (%s): ", default=identity.email)
            password = StdIO.prompt("Enter password (RETURN to keep existing): ", default=None)

            if not Datum.is_email_address(email):
                logger.error("The email address '%s' is not valid." % email)
                exit(1)

            if password and not CognitoUserIdentity.is_valid_password(password):
                logger.error("The password '%s' is not valid." % password)
                exit(1)

            report = CognitoUserIdentity(identity.username, None, None, None, email, given_name, family_name, password)

            auth = gatekeeper.login(credentials)                          # renew credentials
            manager = CognitoUpdateManager(requests, auth.id_token)
            manager.update(report)

            # update credentials...
            credentials.email = report.email
            if password:
                credentials.password = report.password

            credentials.save(Host, encryption_key=credentials.password)


    # ----------------------------------------------------------------------------------------------------------------
    # end...

        if report is not None:
            print(JSONify.dumps(report, indent=cmd.indent))

    except KeyboardInterrupt:
        print(file=sys.stderr)

    except HTTPConflictException as ex:
        logger.error("the email address '%s' is already in use." % report.email)
        exit(1)



#!/usr/bin/env python3

"""
Created on 16 Mar 2017

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

source repo: scs_analysis

DESCRIPTION
The sample_interval utility is used to discover the jitter present in a data stream.

Input data must in the form of a JSON document. A command parameter specifies the path to the node within
the document that is to be measured. The node must represent time as an ISO 8601 localised date / time. The output of
the sample_interval utility includes the date / time value, and the time difference from the preceding value
in seconds.

SYNOPSIS
sample_interval.py [-p PRECISION] [-v] [PATH]

EXAMPLES
osio_topic_history.py -m1 /orgs/south-coast-science-demo/brighton/loc/1/gases | sample_interval.py -p3 rec

DOCUMENT EXAMPLE - INPUT
{"tag": "scs-bgx-401", "rec": "2018-03-27T09:54:41.042+00:00", "val": {
"NO2": {"weV": 0.29563, "aeV": 0.280879, "weC": 0.009569, "cnc": 61.0},
"Ox": {"weV": 0.406819, "aeV": 0.387443, "weC": -0.010706, "cnc": 34.7},
"NO": {"weV": 0.319692, "aeV": 0.292129, "weC": 0.028952, "cnc": 165.5},
"CO": {"weV": 0.395819, "aeV": 0.289317, "weC": 0.113108, "cnc": 311.3},
"sht": {"hmd": 82.4, "tmp": 12.6}}}

DOCUMENT EXAMPLE - OUTPUT
{"time": "2018-03-27T10:14:51.028+00:00", "diff": 9.987}
"""

import sys

from scs_analysis.cmd.cmd_sample_interval import CmdSampleInterval

from scs_core.data.datetime import LocalizedDatetime
from scs_core.data.interval import Interval
from scs_core.data.json import JSONify
from scs_core.data.path_dict import PathDict


# --------------------------------------------------------------------------------------------------------------------

if __name__ == '__main__':

    document_count = 0
    processed_count = 0

    # ----------------------------------------------------------------------------------------------------------------
    # cmd...

    cmd = CmdSampleInterval()

    if cmd.verbose:
        print("sample_interval: %s" % cmd, file=sys.stderr)
        sys.stderr.flush()


    try:
        # ------------------------------------------------------------------------------------------------------------
        # run...

        prev_time = None

        for line in sys.stdin:
            datum = PathDict.construct_from_jstr(line)

            if datum is None:
                break

            document_count += 1

            time = LocalizedDatetime.construct_from_iso8601(datum.node(cmd.path))

            interval = Interval.construct(prev_time, time, cmd.precision)
            print(JSONify.dumps(interval))
            sys.stdout.flush()

            prev_time = time

            processed_count += 1


    # ----------------------------------------------------------------------------------------------------------------
    # end...

    except KeyError as ex:
        print("sample_interval: KeyError: %s" % ex, file=sys.stderr)

    except KeyboardInterrupt:
        print(file=sys.stderr)

    finally:
        if cmd.verbose:
            print("sample_interval: documents: %d processed: %d" % (document_count, processed_count), file=sys.stderr)



#!/usr/bin/env python3

"""
Created on 2 Nov 2019

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

source repo: scs_analysis

DESCRIPTION
The csv_collation_summary utility is used alongside csv_collator to report on the effect of one independent variable
on one or more dependent variables, for a range of independent variable deltas.

Input is gained from a sequence of CSV files, as created by the csv_collator utility. Input rows with missing or
malformed values for any of the variables are ignored. If the specified columns for any of the named variables
are missing, the utility terminates.

The output of csv_collation_summary is a sequence of JSON documents, detailing the min, median and max values for the
independent variable, with the median and standard deviation of each of the dependent variables.

If the --verbose flag is used, a summary of the data processed is written to stderr.

SYNOPSIS
csv_collation_summary.py  -f FILE_PREFIX -i IND_PATH [-p IND_PRECISION DEP_PRECISION] [-v] DEP_PATH_1 .. DEP_PATH_N

EXAMPLES
csv_collation_summary.py -v -f collated_5rH/joined_PM_meteo_data_2019-02_2019-07_15min_ \
-i th.praxis.val.hmd pm1_scaling pm2p5_scaling pm10_scaling | \
csv_writer.py -v collated_5rH/summary.csv

FILES
Input file names must be of the form:
FILE-PREFIX_DOMAIN-LOW_DOMAIN-HIGH.csv

DOCUMENT EXAMPLE - OUTPUT
{"domain": "70.0 - 75.0", "praxis": {"climate": {"val": {"hmd": {"min": 70.0, "avg": 72.6, "max": 74.9}}}},
"error": {"pm1": {"avg": 2.648, "stdev": 2.458}, "pm2p5": {"avg": 2.992, "stdev": 2.652},
"pm10": {"avg": 2.77, "stdev": 2.456}}, "samples": 2307}

SEE ALSO
scs_analysis/csv_collator

RESOURCES
https://en.wikipedia.org/wiki/Dependent_and_independent_variables
"""

import sys

from scs_analysis.cmd.cmd_csv_collation_summary import CmdCollationSummary
from scs_analysis.handler.csv_collator import CSVCollatorBin

from scs_core.csv.csv_reader import CSVReader, CSVReaderException

from scs_core.data.json import JSONify
from scs_core.data.model_delta import ModelDelta
from scs_core.data.path_dict import PathDict


# --------------------------------------------------------------------------------------------------------------------

if __name__ == '__main__':

    file_count = 0
    total_rows = 0
    total_processed = 0

    reader = None
    domain = None

    # ----------------------------------------------------------------------------------------------------------------
    # cmd...

    cmd = CmdCollationSummary()

    if not cmd.is_valid():
        cmd.print_help(sys.stderr)
        exit(2)

    if cmd.verbose:
        print("csv_collation_summary: %s" % cmd, file=sys.stderr)

    try:
        for filename in cmd.filenames:

            file_count += 1
            rows = 0
            processed = 0

            # --------------------------------------------------------------------------------------------------------
            # resources...

            try:
                domain_min, domain_max = CSVCollatorBin.parse(filename)
            except ValueError:
                print("csv_collation_summary: skipping file: %s" % filename, file=sys.stderr)
                sys.stderr.flush()
                continue

            try:
                reader = CSVReader.construct_for_file(filename)
            except FileNotFoundError:
                print("csv_collation_summary: file not found: %s" % filename, file=sys.stderr)
                exit(1)

            if cmd.verbose:
                print("csv_collation_summary: %s" % filename, file=sys.stderr)
                sys.stderr.flush()


            # --------------------------------------------------------------------------------------------------------
            # run...

            delta = ModelDelta.construct(domain_min, domain_max, cmd.ind_path, cmd.ind_prec, cmd.dep_paths,
                                         cmd.dep_prec)

            try:
                for row in reader.rows():
                    datum = PathDict.construct_from_jstr(row)
                    paths = datum.paths()

                    rows += 1

                    if cmd.ind_path not in paths:
                        print("csv_collation_summary: ind_path not in datum: %s" % cmd.ind_path, file=sys.stderr)
                        exit(1)

                    try:
                        ind_value = float(datum.node(cmd.ind_path))
                    except ValueError:
                        continue                        # independent value is NaN - skip this row

                    dependents = {}
                    for dep_path in cmd.dep_paths:
                        if dep_path not in paths:
                            print("csv_collation_summary: dep_path not in datum: %s" % dep_path, file=sys.stderr)
                            exit(1)

                        try:
                            dep_value = float(datum.node(dep_path))
                        except ValueError:
                            break

                        dependents[dep_path] = dep_value

                    try:
                        delta.append(ind_value, dependents)
                    except IndexError:
                        continue                        # at least one dependent value is NaN - skip this row

                    processed += 1

            except CSVReaderException as ex:
                if cmd.verbose:
                    print("csv_collation_summary: terminating: %s" % ex, file=sys.stderr)
                    exit(1)

            finally:
                if reader is not None:
                    reader.close()

            print(JSONify.dumps(delta))
            sys.stdout.flush()

            if cmd.verbose:
                print("csv_collation_summary: rows: %d processed: %d" % (rows, processed), file=sys.stderr)

            total_rows += rows
            total_processed += processed


    # ----------------------------------------------------------------------------------------------------------------
    # end...

    except KeyError as ex:
        print("csv_collation_summary: KeyError: %s" % ex, file=sys.stderr)

    except KeyboardInterrupt:
        print(file=sys.stderr)

    finally:
        if cmd and cmd.verbose:
            print("csv_collation_summary: total files: %d rows: %d processed: %d" %
                  (file_count, total_rows, total_processed), file=sys.stderr)



#!/usr/bin/env python3

"""
Created on 22 Feb 2019

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

source repo: scs_analysis

DESCRIPTION
The csv_join utility performs an SQL join operation on a pair of CSV files. This is particularly useful where a
comparison is to be made on timeline data from two devices.

All four join types are supported: INNER, LEFT outer, RIGHT outer and FULL outer. Inner joins are used when the output
should only contain rows that match on their primary key.

Output is in the form of a sequence of JSON documents containing three fields: rec, left table contents (labelled for
the left PREFIX) and right table contents (labelled for the right PREFIX). The left and right contents do not contain
their primary key fields. For outer joins, content nodes may be null.

The --iso8601 flag is provided to indicate that the primary key should be interpreted as a ISO 8601 datetime. This is
useful where data sets use alternate datetime formats such as 2019-02-22T01:00:00Z and 2019-02-22T01:00:00+00:00.

If the --verbose flag is used, a summary of the join operation is written to stderr.

Warning: if LEFT, RIGHT or FULL joins are used, the first document written to stdout may not contain all of the fields
of the input CSV pair. In these cases, if the output is piped to the csv_writer utility, then the csv_writer must be
used in header-scan mode.

SYNOPSIS
csv_join.py [-t TYPE] -l PREFIX PK FILENAME -r PREFIX PK FILENAME [-i] [-v]

EXAMPLES
csv_join.py -i -v -l praxis rec praxis_301/praxis_301_2018-08.csv -r ref rec ref/ref_2018-08.csv

DOCUMENT EXAMPLE - INPUT
left:
{"rec": "2019-02-01T02:00:00Z", "val": {"NO2": {"weV": 0.297185, "cnc": 40.8, "aeV": 0.298467, "weC": 0.002271}}}

right:
{"rec": "2019-02-01T02:00:00Z", "val": {"NO2": {"status": "P", "units": "ugm-3", "dns": 34.0}}}

DOCUMENT EXAMPLE - OUTPUT
{"rec": "2019-02-01T02:00:00Z", "praxis": {"val": {"NO2": {"weV": 0.297185, "cnc": 40.8, "aeV": 0.298467,
"weC": 0.002271}}}, "ref": {"val": {"NO2": {"status": "P", "units": "ugm-3", "dns": 34.0}}}}

SEE ALSO
scs_analysis/csv_reader
scs_analysis/csv_writer

RESOURCES
https://www.w3schools.com/sql/sql_join.asp
"""

import sys

from scs_analysis.cmd.cmd_csv_join import CmdCSVJoin

from scs_core.csv.csv_reader import CSVReader

from scs_core.data.join import Join
from scs_core.data.json import JSONify
from scs_core.data.path_dict import PathDict


# --------------------------------------------------------------------------------------------------------------------

if __name__ == '__main__':

    reader = None
    result = None

    left_document_count = 0
    left_processed_count = 0

    right_document_count = 0
    right_processed_count = 0

    joined_count = 0

    # ----------------------------------------------------------------------------------------------------------------
    # cmd...

    cmd = CmdCSVJoin()

    if not cmd.is_valid():
        cmd.print_help(sys.stderr)
        exit(2)

    if not Join.is_valid_type(cmd.type):
        print("csv_join: invalid join type: %s" % cmd.type, file=sys.stderr)
        exit(2)

    if cmd.verbose:
        print("csv_join: %s" % cmd, file=sys.stderr)

    try:
        # ------------------------------------------------------------------------------------------------------------
        # resources...

        join = Join.construct(cmd.left_prefix, cmd.left_pk, cmd.right_prefix, cmd.right_pk, cmd.iso8601)

        # read left CSV file...
        try:
            reader = CSVReader.construct_for_file(cmd.left_filename)
        except FileNotFoundError:
            print("csv_join: file not found: %s" % cmd.left_filename, file=sys.stderr)
            exit(1)

        for row in reader.rows():
            jstr = row.strip()
            datum = PathDict.construct_from_jstr(row)

            if datum is None:
                continue

            left_document_count += 1

            if cmd.left_pk not in datum.paths():
                print("csv_join: pk '%s' missing: %s" % (cmd.left_pk, jstr), file=sys.stderr)
                exit(1)

            if datum.node(cmd.left_pk) == '':
                continue

            try:
                join.append_to_left(datum)
            except ValueError as ex:
                print("csv_join: invalid pk '%s' in: %s" % (datum.node(cmd.left_pk), jstr), file=sys.stderr)
                exit(1)

            left_processed_count += 1

        reader.close()

        # read right CSV file...
        try:
            reader = CSVReader.construct_for_file(cmd.right_filename)
        except FileNotFoundError:
            print("csv_join: file not found: %s" % cmd.right_filename, file=sys.stderr)
            exit(1)

        for row in reader.rows():
            jstr = row.strip()
            datum = PathDict.construct_from_jstr(row)

            if datum is None:
                continue

            right_document_count += 1

            if cmd.right_pk not in datum.paths():
                print("csv_join: pk '%s' missing: %s" % (cmd.right_pk, jstr), file=sys.stderr)
                exit(1)

            if datum.node(cmd.right_pk) == '':
                continue

            try:
                join.append_to_right(datum)
            except (TypeError, ValueError) as ex:
                print("csv_join: invalid pk '%s' in: %s" % (datum.node(cmd.right_pk), jstr), file=sys.stderr)
                exit(1)

            right_processed_count += 1

        reader.close()

        if cmd.verbose:
            print("csv_join: %s" % join, file=sys.stderr)
            sys.stderr.flush()


        # ------------------------------------------------------------------------------------------------------------
        # run...

        if cmd.type == 'LEFT':
            operation = join.left

        elif cmd.type == 'RIGHT':
            operation = join.right

        elif cmd.type == 'FULL':
            operation = join.full

        else:
            operation = join.inner

        for datum in operation():
            print(JSONify.dumps(datum))
            sys.stdout.flush()

            joined_count += 1


    # ----------------------------------------------------------------------------------------------------------------
    # end...

    except KeyError as ex:
        print("csv_join: KeyError: %s" % ex, file=sys.stderr)

    except KeyboardInterrupt:
        print(file=sys.stderr)

    finally:
        if cmd.verbose:
            print("csv_join: left: documents: %d processed: %d" % (left_document_count, left_processed_count),
                  file=sys.stderr)
            print("csv_join: right: documents: %d processed: %d" % (right_document_count, right_processed_count),
                  file=sys.stderr)
            print("csv_join: joined: %d" % joined_count,
                  file=sys.stderr)



"""
Created on 22 Nov 2021

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)
"""

import optparse


# --------------------------------------------------------------------------------------------------------------------

class CmdCognitoCredentials(object):
    """unix command line handler"""

    def __init__(self):
        """
        Constructor
        """
        self.__parser = optparse.OptionParser(usage="%prog [-c CREDENTIALS] [{ -s | -t | -d }] [-v]",
                                              version="%prog 1.0")

        # identity...
        self.__parser.add_option("--credentials", "-c", type="string", action="store", dest="credentials_name",
                                 help="the stored credentials to be presented")

        # operations...
        self.__parser.add_option("--set", "-s", action="store_true", dest="set", default=False,
                                 help="set the credentials")

        self.__parser.add_option("--test", "-t", action="store_true", dest="test", default=False,
                                 help="test the credentials")

        self.__parser.add_option("--delete", "-d", action="store_true", dest="delete", default=False,
                                 help="delete the credentials")

        # output...
        self.__parser.add_option("--verbose", "-v", action="store_true", dest="verbose", default=False,
                                 help="report narrative to stderr")

        self.__opts, self.__args = self.__parser.parse_args()


    # ----------------------------------------------------------------------------------------------------------------

    def is_valid(self):
        count = 0

        if self.set:
            count += 1

        if self.test:
            count += 1

        if self.delete:
            count += 1

        if count > 1:
            return False

        return True


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def credentials_name(self):
        return self.__opts.credentials_name


    @property
    def set(self):
        return self.__opts.set


    @property
    def test(self):
        return self.__opts.test


    @property
    def delete(self):
        return self.__opts.delete


    @property
    def verbose(self):
        return self.__opts.verbose


    # ----------------------------------------------------------------------------------------------------------------

    def print_help(self, file):
        self.__parser.print_help(file)


    def __str__(self, *args, **kwargs):
        return "CmdCognitoCredentials:{credentials_name:%s, set:%s, test:%s, delete:%s, verbose:%s}" % \
               (self.credentials_name, self.set, self.test, self.delete, self.verbose)



"""
Created on 27 Sep 2018

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)
"""

import json
import sys

from scs_core.data.datetime import LocalizedDatetime
from scs_core.data.json import JSONify
from scs_core.data.publication import Publication, ReceivedPublication


# --------------------------------------------------------------------------------------------------------------------
# subscription handler...

class AWSMQTTSubscriptionHandler(object):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, reporter, comms=None, wrap=True, timed=False, echo=False):
        """
        Constructor
        """
        self.__reporter = reporter
        self.__comms = comms
        self.__wrap = wrap
        self.__timed = timed
        self.__echo = echo


    # ----------------------------------------------------------------------------------------------------------------

    def handle(self, _client, _userdata, message):
        payload = message.payload.decode()
        payload_jdict = json.loads(payload)

        publication = Publication(message.topic, payload_jdict) if self.__wrap else payload_jdict

        if self.__timed:
            publication = ReceivedPublication(LocalizedDatetime.now(), publication)

        try:
            self.__comms.connect()
            self.__comms.write(JSONify.dumps(publication), False)

        except ConnectionError:
            try:
                self.__reporter.print("handle: ConnectionError: %s." % self.__comms.address)
            except AttributeError:
                self.__reporter.print("handle: ConnectionError.")
                exit(0)

        finally:
            self.__comms.close()

        if self.__echo:
            print(JSONify.dumps(publication))
            sys.stdout.flush()

            self.__reporter.print("received: %s" % JSONify.dumps(publication))


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "AWSMQTTSubscriptionHandler:{reporter:%s, comms:%s, wrap:%s, timed:%s, echo:%s}" % \
               (self.__reporter, self.__comms, self.__wrap, self.__timed, self.__echo)





"""
Created on 9 Mar 2019

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)
"""

import optparse


# --------------------------------------------------------------------------------------------------------------------

class CmdMQTTPeers(object):
    """unix command line handler"""

    def __init__(self):
        """
        Constructor
        """
        self.__parser = optparse.OptionParser(usage="%prog [-a] { -p [-e] | -l [-n HOSTNAME] [-t TOPIC] | -m | "
                                                    "-c HOSTNAME TAG SHARED_SECRET TOPIC | "
                                                    "-u HOSTNAME { -s SHARED_SECRET | -t TOPIC } | "
                                                    "-r HOSTNAME } [-i INDENT] [-v]", version="%prog 1.0")

        # source...
        self.__parser.add_option("--aws", "-a", action="store_true", dest="aws", default=False,
                                 help="Use AWS S3 instead of local storage")

        # functions...
        self.__parser.add_option("--import", "-p", action="store_true", dest="import_peers", default=False,
                                 help="import MQTT peers from stdin")

        self.__parser.add_option("--list", "-l", action="store_true", dest="list", default=False,
                                 help="list the stored MQTT peers to stdout")

        self.__parser.add_option("--missing", "-m", action="store_true", dest="missing", default=False,
                                 help="list known devices missing from S3 MQTT peers")

        self.__parser.add_option("--create", "-c", type="string", nargs=4, action="store", dest="create",
                                 help="create an MQTT peer")

        self.__parser.add_option("--update", "-u", type="string", nargs=1, action="store", dest="update",
                                 help="update an MQTT peer")

        self.__parser.add_option("--remove", "-r", type="string", nargs=1, action="store", dest="remove",
                                 help="delete an MQTT peer")

        # filters...
        self.__parser.add_option("--hostname", "-n", type="string", nargs=1, action="store", dest="hostname",
                                 help="filter peers with the given hostname substring")

        self.__parser.add_option("--shared-secret", "-s", type="string", nargs=1, action="store", dest="shared_secret",
                                 help="specify shared secret")

        self.__parser.add_option("--topic", "-t", type="string", nargs=1, action="store", dest="topic",
                                 help="specify topic")

        # output...
        self.__parser.add_option("--echo", "-e", action="store_true", dest="echo", default=False,
                                 help="echo stdin to stdout (import only)")

        self.__parser.add_option("--indent", "-i", type="int", nargs=1, action="store", dest="indent",
                                 help="pretty-print the output with INDENT (not with echo)")

        self.__parser.add_option("--verbose", "-v", action="store_true", dest="verbose", default=False,
                                 help="report narrative to stderr")

        self.__opts, self.__args = self.__parser.parse_args()


    # ----------------------------------------------------------------------------------------------------------------

    def is_valid(self):
        count = 0

        if self.is_import():
            count += 1

        if self.missing:
            count += 1

        if self.list:
            count += 1

        if self.is_create():
            count += 1

        if self.is_update():
            count += 1

        if self.is_remove():
            count += 1

        if count != 1:
            return False

        if not self.is_import() and self.echo:
            return False

        if self.__opts.list is None and (self.__opts.hostname is not None or self.__opts.for_topic is not None):
            return False

        if self.missing and not self.aws:
            return False

        if self.echo and self.indent is not None:
            return False

        if self.is_update() and not self.shared_secret and not self.topic:
            return False

        return True


    # ----------------------------------------------------------------------------------------------------------------

    def is_import(self):
        return self.__opts.import_peers


    def is_create(self):
        return self.__opts.create is not None


    def is_update(self):
        return self.__opts.update is not None


    def is_remove(self):
        return self.__opts.remove is not None


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def list(self):
        return self.__opts.list


    @property
    def missing(self):
        return self.__opts.missing


    @property
    def create_hostname(self):
        return None if self.__opts.create is None else self.__opts.create[0]


    @property
    def create_tag(self):
        return None if self.__opts.create is None else self.__opts.create[1]


    @property
    def create_shared_secret(self):
        return None if self.__opts.create is None else self.__opts.create[2]


    @property
    def create_topic(self):
        return None if self.__opts.create is None else self.__opts.create[3]


    @property
    def update_hostname(self):
        return self.__opts.update


    @property
    def remove_hostname(self):
        return self.__opts.remove


    @property
    def hostname(self):
        return self.__opts.hostname


    @property
    def shared_secret(self):
        return self.__opts.shared_secret


    @property
    def topic(self):
        return self.__opts.topic


    @property
    def aws(self):
        return self.__opts.aws


    @property
    def echo(self):
        return self.__opts.echo


    @property
    def indent(self):
        return self.__opts.indent


    @property
    def verbose(self):
        return self.__opts.verbose


    # ----------------------------------------------------------------------------------------------------------------

    def print_help(self, file):
        self.__parser.print_help(file)

    def __str__(self, *args, **kwargs):
        return "CmdMQTTPeers:{import:%s, list:%s, missing:%s, create:%s, update:%s, " \
               "remove:%s, hostname:%s, shared_secret:%s, topic:%s, aws:%s, echo:%s, indent:%s, " \
               "verbose:%s}" %  \
               (self.__opts.import_peers, self.list, self.missing, self.__opts.create, self.__opts.update,
                self.__opts.remove, self.hostname, self.shared_secret, self.topic, self.aws, self.echo, self.indent,
                self.verbose)



#!/usr/bin/env python3

"""
Created on 11 Feb 2018

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

source repo: scs_analysis

DESCRIPTION
The timer utility is used to report elapsed time for a command line job. To use, it should be threaded into a pipeline,
where it passes stdin to stout. On terminating, the timer utility reports a timedelta to stderr.

SYNOPSIS
timer.py

EXAMPLES
csv_reader.py source*.csv | timer.py | csv_writer.py target.csv

https://stackoverflow.com/questions/26692284/how-to-prevent-brokenpipeerror-when-doing-a-flush-in-python
"""

import sys
import time

from scs_core.data.json import JSONify
from scs_core.data.timedelta import Timedelta


# --------------------------------------------------------------------------------------------------------------------

if __name__ == '__main__':

    # ----------------------------------------------------------------------------------------------------------------
    # resources...

    print("timer: starting", file=sys.stderr)
    sys.stderr.flush()

    start_time = time.time()

    try:
        # ------------------------------------------------------------------------------------------------------------
        # run...

        for line in sys.stdin:
            print(line, end='')
            sys.stdout.flush()

    # ----------------------------------------------------------------------------------------------------------------
    # end...

    except KeyboardInterrupt:
        print(file=sys.stderr)

    except (BrokenPipeError, IOError):
        pass

    # ----------------------------------------------------------------------------------------------------------------
    # close...

    finally:
        elapsed_time = time.time() - start_time
        delta = Timedelta(seconds=elapsed_time)

        print("timer: %s" % JSONify.dumps(delta.as_json()).strip('"'), file=sys.stderr)

        sys.stderr.close()



#!/usr/bin/env python3

"""
Created on 23 Oct 2020

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

source repo: scs_analysis

DESCRIPTION
The sample_slope utility it used to find the slope for a given field in a sequence of JSON documents. Slope is
defined as change in value / change in time (in seconds).

The output document replaced the input field with field.cur (the current value) and field.slope. For the initial
input document, the corresponding output document's field.slope value is null. A --tally flag specifies the number
of rolling samples in the regression.

SYNOPSIS
sample_slope.py [-i ISO] [-t TALLY] [-p PRECISION] [-v] PATH

EXAMPLES
csv_reader.py -v ~/climate.csv | sample_slope.py -v val.tmp

DOCUMENT EXAMPLE - INPUT
{"val": {"hmd": 67.6, "tmp": 20.5, "bar": ""}, "rec": "2020-10-23T12:19:20Z", "tag": "scs-bgx-401"}

DOCUMENT EXAMPLE - OUTPUT
{"val": {"hmd": 67.6, "tmp": {"cur": 20.5, "slope": -0.001667}, "bar": ""}, "rec": "2020-10-23T12:19:20Z",
"tag": "scs-bgx-401"}
"""

import sys

from scs_analysis.cmd.cmd_sample_slope import CmdSampleSlope

from scs_core.data.datetime import LocalizedDatetime
from scs_core.data.datum import Datum
from scs_core.data.json import JSONify
from scs_core.data.linear_regression import LinearRegression
from scs_core.data.path_dict import PathDict


# --------------------------------------------------------------------------------------------------------------------

if __name__ == '__main__':

    document_count = 0
    processed_count = 0

    # ----------------------------------------------------------------------------------------------------------------
    # cmd...

    cmd = CmdSampleSlope()

    if not cmd.is_valid():
        cmd.print_help(sys.stderr)
        exit(2)

    if cmd.verbose:
        print("sample_slope: %s" % cmd, file=sys.stderr)

    try:
        # ------------------------------------------------------------------------------------------------------------
        # resources...

        regression = LinearRegression(tally=cmd.tally)


        # ------------------------------------------------------------------------------------------------------------
        # run...

        prev_rec = None
        prev_value = None

        for line in sys.stdin:
            jstr = line.strip()
            datum = PathDict.construct_from_jstr(jstr)

            if datum is None:
                continue

            document_count += 1

            paths = datum.paths()

            # rec...
            if cmd.iso not in paths:
                print("sample_slope: field %s not present." % cmd.iso, file=sys.stderr)
                exit(1)

            rec_node = datum.node(cmd.iso)
            rec = LocalizedDatetime.construct_from_iso8601(rec_node)

            if rec is None:
                print("sample_slope: invalid ISO 8601 value '%s' in %s." % (rec_node, jstr), file=sys.stderr)
                exit(1)

            # value...
            if cmd.path not in paths:
                print("sample_slope: field %s not present in %s." % (cmd.path, jstr), file=sys.stderr)
                exit(1)

            value_node = datum.node(cmd.path)
            value = Datum.float(value_node)

            if value is None:
                print("sample_slope: invalid numeric value '%s' in %s." % (value_node, jstr), file=sys.stderr)
                exit(1)

            # slope...
            regression.append(rec, value)

            if regression.has_regression():
                m, _ = regression.line()
                slope = None if m is None else round(m, cmd.precision)

            else:
                slope = None

            # target...
            target = PathDict()

            for path in datum.paths():
                if path == cmd.path:
                    target.append('.'.join((cmd.path, 'cur')), value)
                    target.append('.'.join((cmd.path, 'slope')), slope)
                    continue

                target.copy(datum, path)

            print(JSONify.dumps(target))
            sys.stdout.flush()

            prev_rec = rec
            prev_value = value
            processed_count += 1


    # ----------------------------------------------------------------------------------------------------------------
    # end...

    except KeyboardInterrupt:
        print(file=sys.stderr)

    finally:
        if cmd.verbose:
            print("sample_slope: documents: %d processed: %d" % (document_count, processed_count), file=sys.stderr)



#!/usr/bin/env python3

"""
Created on 24 Jan 2022

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

source repo: scs_analysis

DESCRIPTION
The cognito_users utility is used to create, update and retrieve AWS Cognito identities. This utility can only be used
by organisation administrators and superusers.

If the --Create function is used, an email is sent to the new user. The verification link in the email must be
excercised in order for the account to gain a CONFIRMED status.

SYNOPSIS
cognito_users.py  [-c CREDENTIALS] { -F [{ -e EMAIL_ADDR | -l ORG_LABEL | -c CONFIRMATION | -s { 1 | 0 } } }] | \
-C | -U -e EMAIL_ADDR | -D -e EMAIL_ADDR } [-i INDENT] [-v]

EXAMPLES
cognito_users.py -Fe bruno.beloff@southcoastscience.com

DOCUMENT EXAMPLE
{"username": "8", "creation-date": "2021-11-24T12:51:12Z", "confirmation-status": "CONFIRMED", "enabled": true,
"email": "bruno.beloff@southcoastscience.com", "given-name": "Bruno", "family-name": "Beloff", "is-super": true}

SEE ALSO
scs_analysis/cognito_credentials

RESOURCES
https://docs.aws.amazon.com/cognito/latest/developerguide/signing-up-users-in-your-app.html
https://docs.aws.amazon.com/cognito/latest/developerguide/user-pool-settings-policies.html
"""

import requests
import sys

from scs_analysis.cmd.cmd_cognito_users import CmdCognitoUsers

from scs_core.aws.security.cognito_account_manager import CognitoCreateManager, CognitoUpdateManager, \
    CognitoDeleteManager

from scs_core.aws.security.cognito_finder import CognitoFinder
from scs_core.aws.security.cognito_login_manager import CognitoLoginManager
from scs_core.aws.security.cognito_user import CognitoUserCredentials, CognitoUserIdentity

from scs_core.aws.security.organisation_manager import OrganisationManager

from scs_core.data.datum import Datum
from scs_core.data.json import JSONify

from scs_core.sys.http_exception import HTTPException, HTTPConflictException
from scs_core.sys.logging import Logging

from scs_host.sys.host import Host


# --------------------------------------------------------------------------------------------------------------------

if __name__ == '__main__':

    logger = None
    gatekeeper = None
    credentials = None
    auth = None
    finder = None
    manager = None
    org = None
    report = None

    try:
        # ------------------------------------------------------------------------------------------------------------
        # cmd...

        cmd = CmdCognitoUsers()

        if not cmd.is_valid():
            cmd.print_help(sys.stderr)
            exit(2)

        Logging.config('cognito_users', verbose=cmd.verbose)
        logger = Logging.getLogger()

        logger.info(cmd)


        # ------------------------------------------------------------------------------------------------------------
        # auth...

        if not cmd.create:
            gatekeeper = CognitoLoginManager(requests)

            # CognitoUserCredentials...
            if not CognitoUserCredentials.exists(Host, name=cmd.credentials_name):
                logger.error("Cognito credentials not available.")
                exit(1)

            try:
                password = CognitoUserCredentials.password_from_user()
                credentials = CognitoUserCredentials.load(Host, name=cmd.credentials_name, encryption_key=password)
            except (KeyError, ValueError):
                logger.error("incorrect password")
                exit(1)

            try:
                auth = gatekeeper.login(credentials)

            except HTTPException as ex:
                logger.error(ex.data)
                exit(1)


        # ------------------------------------------------------------------------------------------------------------
        # resources...

        if not cmd.create:
            finder = CognitoFinder(requests)

        if cmd.org_label:
            manager = OrganisationManager(requests)
            org = manager.get_organisation_by_label(auth.id_token, cmd.org_label)

            if org is None:
                logger.error("no organisation found for label: '%s'." % cmd.org_label)
                exit(1)


        # ------------------------------------------------------------------------------------------------------------
        # run...

        if cmd.find:
            if cmd.email is not None:
                report = sorted(finder.find_by_email(auth.id_token, cmd.email))

            elif cmd.org_label is not None:
                org_users = manager.find_users_by_organisation(auth.id_token, org.org_id)
                usernames = [org_user.username for org_user in org_users]

                report = sorted(finder.find_by_usernames(auth.id_token, usernames))

            elif cmd.confirmation_status is not None:
                report = sorted(finder.find_by_status(auth.id_token,
                                                      CognitoUserIdentity.status(cmd.confirmation_status)))

            elif cmd.enabled is not None:
                report = sorted(finder.find_by_enabled(auth.id_token, cmd.enabled))

            else:
                report = sorted(finder.find_all(auth.id_token))

        if cmd.create:
            # create...
            if not Datum.is_email_address(cmd.email):
                logger.error("The email address '%s' is not valid." % cmd.email)
                exit(1)

            # if not CognitoUserIdentity.is_valid_password(password):
            #     logger.error("The password '%s' is not valid." % password)
            #     exit(1)

            identity = CognitoUserIdentity(None, None, None, None,
                                           cmd.email, cmd.given_name, cmd.family_name, None)

            manager = CognitoCreateManager(requests)
            report = manager.create(identity)

        if cmd.update:
            # find...
            identity = finder.get_by_email(auth.id_token, cmd.update)

            if identity is None:
                logger.error("no identity found for email: '%s'." % cmd.update)
                exit(1)

            # update...
            enabled = identity.enabled if cmd.enabled is None else cmd.enabled
            email = identity.email if cmd.email is None else cmd.email
            given_name = identity.given_name if cmd.given_name is None else cmd.given_name
            family_name = identity.family_name if cmd.family_name is None else cmd.family_name

            if not Datum.is_email_address(email):
                logger.error("The email address '%s' is not valid." % email)
                exit(1)

            # if password and not CognitoUserIdentity.is_valid_password(password):
            #     logger.error("The password '%s' is not valid." % password)
            #     exit(1)

            report = CognitoUserIdentity(identity.username, None, None,
                                         enabled, email, given_name, family_name, None)

            auth = gatekeeper.login(credentials)                          # renew credentials
            manager = CognitoUpdateManager(requests, auth.id_token)
            manager.update(identity)

        if cmd.delete:
            # TODO: delete user from organisations
            manager = CognitoDeleteManager(requests, auth.id_token)
            manager.delete(cmd.delete)


    # ----------------------------------------------------------------------------------------------------------------
    # end...

        if report is not None:
            print(JSONify.dumps(report, indent=cmd.indent))

    except KeyboardInterrupt:
        print(file=sys.stderr)

    except HTTPConflictException as ex:
        logger.error("the email address '%s' is already in use." % report.email)
        exit(1)



#!/usr/bin/env python3

"""
Created on 18 Nov 2016

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

source repo: scs_analysis

DESCRIPTION
The osio_topic_publisher utility is used to prepare data for publication by the osio_mqtt_client script. The
osio_topic_publisher works by taking data from stdin, wrapping it in a JSON document whose only field has the name of
the given topic, and presenting the result on stdout.

Note that the osio_topic_publisher in scs_analysis necessarily works differently to the osio_topic_publisher in scs_dev.
This is because scs_dev version has access to a device project specification, and therefore can find the topic path
automatically. For the scs_analysis version, the full topic path should be given explicitly.

SYNOPSIS
osio_topic_publisher.py -t TOPIC [-o] [-v]

EXAMPLES
osio_topic_publisher.py -t /users/southcoastscience-dev/test/json

SEE ALSO
scs_analysis/osio_mqtt_client

RESOURCES
https://opensensorsio.helpscoutdocs.com/article/84-overriding-timestamp-information-in-message-payload
"""

import json
import sys

from collections import OrderedDict

from scs_analysis.cmd.cmd_osio_topic_publisher import CmdOSIOTopicPublisher

from scs_core.data.json import JSONify
from scs_core.data.publication import Publication

from scs_core.osio.config.project import Project

from scs_core.sys.system_id import SystemID

from scs_host.sys.host import Host


# TODO: remove channel handling

# --------------------------------------------------------------------------------------------------------------------

if __name__ == '__main__':

    # ----------------------------------------------------------------------------------------------------------------
    # cmd...

    cmd = CmdOSIOTopicPublisher()

    if not cmd.is_valid():
        cmd.print_help(sys.stderr)
        exit(2)

    if cmd.verbose:
        print("osio_topic_publisher: %s" % cmd, file=sys.stderr)

    try:
        # ------------------------------------------------------------------------------------------------------------
        # resources...

        # topic...
        if cmd.channel:
            # SystemID...
            system_id = SystemID.load(Host)

            if system_id is None:
                print("osio_topic_publisher: SystemID not available.", file=sys.stderr)
                exit(1)

            if cmd.verbose:
                print("osio_topic_publisher: %s" % system_id, file=sys.stderr)

            project = Project.load(Host)

            if project is None:
                print("osio_topic_publisher: Project not available.", file=sys.stderr)
                exit(1)

            topic = project.channel_path(cmd.channel, system_id)

        else:
            topic = cmd.topic

        # TODO: check if topic exists

        if cmd.verbose:
            print("osio_topic_publisher: %s" % topic, file=sys.stderr)
            sys.stderr.flush()


        # ------------------------------------------------------------------------------------------------------------
        # run...

        for line in sys.stdin:
            try:
                jdict = json.loads(line)
            except ValueError:
                continue

            if cmd.override:
                payload = OrderedDict({'__timestamp': jdict['rec']})
                payload.update(jdict)

            else:
                payload = jdict

            # time.sleep(1)

            publication = Publication(topic, payload)

            print(JSONify.dumps(publication))
            sys.stdout.flush()


    # ----------------------------------------------------------------------------------------------------------------
    # end...

    except KeyboardInterrupt:
        print(file=sys.stderr)



"""
Created on 20 Feb 2017

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

source repo: scs_analysis
"""

import optparse

from scs_core.data.datetime import LocalizedDatetime


# --------------------------------------------------------------------------------------------------------------------

class CmdOSIOTopicHistory(object):
    """unix command line handler"""

    def __init__(self):
        """
        Constructor
        """
        self.__parser = optparse.OptionParser(usage="%prog { -m MINUTES | -s START [-e END] } [-p SECONDS] [-w] "
                                                    "[-v] PATH", version="%prog 1.0")

        # optional...
        self.__parser.add_option("--minutes", "-m", type="int", nargs=1, action="store", dest="minutes",
                                 help="starting minutes ago")

        self.__parser.add_option("--start", "-s", type="string", nargs=1, action="store", dest="start",
                                 help="ISO 8601 datetime start")

        self.__parser.add_option("--end", "-e", type="string", nargs=1, action="store", dest="end",
                                 help="ISO 8601 datetime end")

        self.__parser.add_option("--pause", "-p", type="float", nargs=1, action="store", dest="pause", default=0.0,
                                 help="pause for SECONDS between retrieved batches (prevents rate limit exceeded)")

        self.__parser.add_option("--wrapping", "-w", action="store_true", dest="include_wrapping", default=False,
                                 help="include message wrapper")

        self.__parser.add_option("--verbose", "-v", action="store_true", dest="verbose", default=False,
                                 help="report narrative to stderr")

        self.__opts, self.__args = self.__parser.parse_args()


    # ----------------------------------------------------------------------------------------------------------------

    def is_valid(self):
        if self.path is None or (self.__opts.start is None and self.minutes is None):
            return False

        if self.__opts.start is not None and LocalizedDatetime.construct_from_iso8601(self.__opts.start) is None:
            return False

        if self.__opts.end is not None and LocalizedDatetime.construct_from_iso8601(self.__opts.end) is None:
            return False

        return True


    # ----------------------------------------------------------------------------------------------------------------

    def use_offset(self):
        return self.minutes is not None


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def minutes(self):
        return self.__opts.minutes


    @property
    def start(self):
        return LocalizedDatetime.construct_from_iso8601(self.__opts.start) if self.__opts.start else None


    @property
    def end(self):
        return LocalizedDatetime.construct_from_iso8601(self.__opts.end) if self.__opts.end else None


    @property
    def pause(self):
        return self.__opts.pause

    @property
    def include_wrapping(self):
        return self.__opts.include_wrapping


    @property
    def verbose(self):
        return self.__opts.verbose


    @property
    def path(self):
        return self.__args[0] if len(self.__args) > 0 else None


    # ----------------------------------------------------------------------------------------------------------------

    def print_help(self, file):
        self.__parser.print_help(file)


    def __str__(self, *args, **kwargs):
        return "CmdOSIOTopicHistory:{minutes:%s, start:%s, end:%s, pause:%s, include_wrapping:%s, " \
               "verbose:%s, path:%s}" % \
                    (self.minutes, self.start, self.end, self.pause, self.include_wrapping,
                     self.verbose, self.path)



"""
Created on 3 Mar 2019

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

source repo: scs_analysis
"""

import optparse


# --------------------------------------------------------------------------------------------------------------------

class CmdSampleDuplicates(object):
    """unix command line handler"""

    def __init__(self):
        """
        Constructor
        """
        self.__parser = optparse.OptionParser(usage="%prog [-c] [-v] PATH", version="%prog 1.0")

        # optional...
        self.__parser.add_option("--counts", "-c", action="store_true", dest="counts", default=False,
                                 help="only list the count of matching documents")

        self.__parser.add_option("--verbose", "-v", action="store_true", dest="verbose", default=False,
                                 help="report narrative to stderr")

        self.__opts, self.__args = self.__parser.parse_args()


    # ----------------------------------------------------------------------------------------------------------------

    def is_valid(self):
        if len(self.__args) != 1:
            return False

        return True


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def counts(self):
        return self.__opts.counts


    @property
    def verbose(self):
        return self.__opts.verbose


    @property
    def path(self):
        return self.__args[0] if len(self.__args) > 0 else None


    # ----------------------------------------------------------------------------------------------------------------

    def print_help(self, file):
        self.__parser.print_help(file)


    def __str__(self, *args, **kwargs):
        return "CmdSampleDuplicates:{counts:%s, verbose:%s, path:%s}" % (self.counts, self.verbose, self.path)



"""
Created on 31 Jul 2021

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)
"""

import optparse


# --------------------------------------------------------------------------------------------------------------------

class CmdBaselineConf(object):
    """
    unix command line handler
    """

    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self):
        self.__parser = optparse.OptionParser(usage="%prog [-a] { -z | -l | -n FROM TO | -c NAME [-t TIMEZONE_NAME] "
                                                    "[-s START] [-e END] [-p AGGREGATION] [-g GAS MINIMUM] [-r GAS] } "
                                                    "[-i INDENT] [-v]", version="%prog 1.0")

        # source...
        self.__parser.add_option("--aws", "-a", action="store_true", dest="aws", default=False,
                                 help="Use AWS S3 instead of local storage for configuration")

        # helpers...
        self.__parser.add_option("--zones", "-z", action="store_true", dest="zones", default=False,
                                 help="list the available timezone names to stderr")

        self.__parser.add_option("--list", "-l", action="store_true", dest="list", default=False,
                                 help="list the available baseline configurations")

        # identity...
        self.__parser.add_option("--duplicate", "-n", type="string", nargs=2, action="store", dest="duplicate",
                                 help="create a new configuration based on FROM")

        self.__parser.add_option("--conf-name", "-c", type="string", nargs=1, action="store", dest="conf_name",
                                 help="the name of the baseline configuration")

        # fields...
        self.__parser.add_option("--timezone", "-t", type="string", nargs=1, action="store", dest="timezone",
                                 help="set the timezone for the tests")

        self.__parser.add_option("--start-hour", "-s", type="int", nargs=1, action="store", dest="start_hour",
                                 help="test starts at the beginning of start hour")

        self.__parser.add_option("--end-hour", "-e", type="int", nargs=1, action="store", dest="end_hour",
                                 help="test ends at the beginning of end hour")

        self.__parser.add_option("--aggregation-period", "-p", type="int", nargs=1, action="store",
                                 dest="aggregation_period", help="aggregation in minutes")

        self.__parser.add_option("--set-gas", "-g", type="string", nargs=2, action="store", dest="set_gas",
                                 help="set minimum for GAS")

        self.__parser.add_option("--remove-gas", "-r", type="string", nargs=1, action="store", dest="remove_gas",
                                 help="remove GAS from minimums")

        # output...
        self.__parser.add_option("--indent", "-i", type="int", nargs=1, action="store", dest="indent",
                                 help="pretty-print the output with INDENT")

        self.__parser.add_option("--verbose", "-v", action="store_true", dest="verbose", default=False,
                                 help="report narrative to stderr")

        self.__opts, self.__args = self.__parser.parse_args()


    # ----------------------------------------------------------------------------------------------------------------

    def is_valid(self):
        count = 0

        if self.zones:
            count += 1

        if self.list:
            count += 1

        if self.duplicate_from is not None:
            count += 1

        if self.conf_name is not None:
            count += 1

        if count != 1:
            return False

        try:
            _ = self.set_gas_minimum
        except ValueError:
            return False

        return True


    def is_complete(self):
        if self.timezone is None or self.start_hour is None or self.end_hour is None or \
                self.aggregation_period is None:
            return False

        return True


    def duplicate(self):
        return self.__opts.duplicate is not None


    def set(self):
        return self.timezone is not None or self.start_hour is not None or self.end_hour is not None or \
               self.aggregation_period is not None or self.__opts.set_gas is not None or self.remove_gas is not None


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def aws(self):
        return self.__opts.aws


    @property
    def zones(self):
        return self.__opts.zones


    @property
    def list(self):
        return self.__opts.list


    @property
    def duplicate_from(self):
        return self.__opts.duplicate[0] if self.__opts.duplicate else None


    @property
    def duplicate_to(self):
        return self.__opts.duplicate[1] if self.__opts.duplicate else None


    @property
    def conf_name(self):
        return self.__opts.conf_name


    @property
    def timezone(self):
        return self.__opts.timezone


    @property
    def start_hour(self):
        return self.__opts.start_hour


    @property
    def end_hour(self):
        return self.__opts.end_hour


    @property
    def aggregation_period(self):
        return self.__opts.aggregation_period


    @property
    def set_gas_name(self):
        return None if self.__opts.set_gas is None else self.__opts.set_gas[0]


    @property
    def set_gas_minimum(self):
        return None if self.__opts.set_gas is None else int(self.__opts.set_gas[1])


    @property
    def remove_gas(self):
        return self.__opts.remove_gas


    @property
    def indent(self):
        return self.__opts.indent


    @property
    def verbose(self):
        return self.__opts.verbose


    # ----------------------------------------------------------------------------------------------------------------

    def print_help(self, file):
        self.__parser.print_help(file)


    def __str__(self, *args, **kwargs):
        return "CmdBaselineConf:{aws:%s, zones:%s, list:%s, list:%s, conf_name:%s, timezone:%s, start_hour:%s, " \
               "end_hour:%s, aggregation_period:%s, set_gas:%s, remove_gas:%s, indent:%s, verbose:%s}" % \
               (self.aws, self.zones, self.list, self.__opts.duplicate, self.conf_name, self.timezone, self.start_hour,
                self.end_hour, self.aggregation_period, self.__opts.set_gas, self.remove_gas, self.indent, self.verbose)



#!/usr/bin/env python3

"""
Created on 18 Feb 2017

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

source repo: scs_analysis

DESCRIPTION
The osio_api_auth utility is used to store or read the API key required by the OpenSensors.io Community Edition
historic data retrieval system.

SYNOPSIS
osio_api_auth.py [-s ORG_ID API_KEY] [-v]

EXAMPLES
osio_api_auth.py -v -s south-coast-science-demo 099add97-6e89-4801-8d12-dd617797cd3b

FILES
~/SCS/osio/osio_api_auth.json

DOCUMENT EXAMPLE
{"org-id": "south-coast-science-demo", "api-key": "099add97-6e89-4801-8d12-dd617797cd3b"}

SEE ALSO
scs_analysis/osio_topic_history
scs_analysis/osio_mqtt_client
scs_analysis/osio_mqtt_control
"""

import sys

from scs_core.data.json import JSONify
from scs_core.osio.client.api_auth import APIAuth

from scs_host.sys.host import Host

from scs_analysis.cmd.cmd_osio_api_auth import CmdOSIOAPIAuth


# --------------------------------------------------------------------------------------------------------------------

if __name__ == '__main__':

    # ----------------------------------------------------------------------------------------------------------------
    # cmd...

    cmd = CmdOSIOAPIAuth()

    if cmd.verbose:
        print("osio_api_auth: %s" % cmd, file=sys.stderr)
        sys.stderr.flush()


    # ----------------------------------------------------------------------------------------------------------------
    # run...

    if cmd.set():
        auth = APIAuth(cmd.org_id, cmd.api_key)

        auth.save(Host)

    else:
        # find self...
        auth = APIAuth.load(Host)

    print(JSONify.dumps(auth))



#!/usr/bin/env python3

"""
Created on 19 Aug 2016

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

source repo: scs_analysis

DESCRIPTION
The sample_noise utility is typically used in a setting where data has a gaussian distribution around a fixed or
slow-moving value. The error analysis shows the difference between the current value, and an exponential
average.

Input data is typically in the form of a JSON document. A command parameter specifies the path to the node within
the document that is to be examined. The node is typically a leaf node integer or float. The output of the
sample_noise utility includes the source value, aggregate, and the error.

SYNOPSIS
sample_noise.py [-p PRECISION] [-v] [PATH]

EXAMPLES
aws_topic_history.py -t 10 /orgs/south-coast-science-demo/brighton/loc/1/gases | sample_noise.py -p 3 val.CO.cnc

DOCUMENT EXAMPLE - INPUT
{"tag": "scs-bgx-401", "rec": "2018-03-27T09:54:41.042+00:00", "val": {
"NO2": {"weV": 0.29563, "aeV": 0.280879, "weC": 0.009569, "cnc": 61.0},
"Ox": {"weV": 0.406819, "aeV": 0.387443, "weC": -0.010706, "cnc": 34.7},
"NO": {"weV": 0.319692, "aeV": 0.292129, "weC": 0.028952, "cnc": 165.5},
"CO": {"weV": 0.395819, "aeV": 0.289317, "weC": 0.113108, "cnc": 311.3},
"sht": {"hmd": 82.4, "tmp": 12.6}}}

DOCUMENT EXAMPLE - OUTPUT
{"rec": "2018-03-27T10:07:11.033+00:00", "val": {"CO": {"cnc": {"src": 263.0, "agr": 194.69, "err": 68.31}}}}
"""

import sys

from scs_analysis.cmd.cmd_sample_filter import CmdSampleFilter

from scs_core.data.json import JSONify
from scs_core.data.path_dict import PathDict


# --------------------------------------------------------------------------------------------------------------------

class SampleNoise(object):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, path, precision):
        """
        Constructor
        """
        self.__path = path
        self.__precision = precision

        self.__aggregate = None


    # ----------------------------------------------------------------------------------------------------------------

    def datum(self, sample):
        latest = float(sample.node(self.__path))

        if self.__aggregate is None:
            self.__aggregate = latest
            return None

        self.__aggregate = (0.9 * self.__aggregate) + (0.1 * latest)
        error = latest - self.__aggregate

        target = PathDict()

        if sample.has_path('rec'):
            target.copy(sample, 'rec')

        target.append(self.__path + '.src', latest)
        target.append(self.__path + '.agr', round(self.__aggregate, self.__precision))
        target.append(self.__path + '.err', round(error, self.__precision))

        return target.node()


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def aggregate(self):
        return self.__aggregate


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        aggregate = "None" if self.__aggregate is None else format(self.__aggregate, '.6f')

        return "SampleNoise:{path:%s, precision:%s, aggregate:%s}" % (self.__path, self.__precision, aggregate)


# --------------------------------------------------------------------------------------------------------------------

if __name__ == '__main__':

    document_count = 0
    processed_count = 0

    # ----------------------------------------------------------------------------------------------------------------
    # cmd...

    cmd = CmdSampleFilter()

    if cmd.verbose:
        print("sample_noise: %s" % cmd, file=sys.stderr)

    try:
        # ------------------------------------------------------------------------------------------------------------
        # resources...

        err = SampleNoise(cmd.path, cmd.precision)

        if cmd.verbose:
            print("sample_noise: %s" % err, file=sys.stderr)
            sys.stderr.flush()


        # ------------------------------------------------------------------------------------------------------------
        # run...

        for line in sys.stdin:
            datum = PathDict.construct_from_jstr(line)

            if datum is None:
                continue

            document_count += 1

            error_datum = err.datum(datum)

            if error_datum is not None:
                print(JSONify.dumps(error_datum))
                sys.stdout.flush()

            processed_count += 1

        if cmd.verbose:
            print(err, file=sys.stderr)


    # ----------------------------------------------------------------------------------------------------------------
    # end...

    except KeyboardInterrupt:
        print(file=sys.stderr)

    finally:
        if cmd.verbose:
            print("sample_noise: documents: %d processed: %d" % (document_count, processed_count), file=sys.stderr)



#!/usr/bin/env python3

"""
Created on 9 Mar 2019

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

DESCRIPTION
The mqtt_peers utility is used to maintain a database of MQTT peers for use with the
scs_analysis/aws_mqtt_control and scs_analysis/osio_mqtt_control utilities.

Options exist to find subsets of authentication documents according to hostname or control topic. Lists of
documents can be exported to or imported from CSV files, enabling the sharing of groups of documents between
users.

Authentication details for specific devices are available on request from South Coast Science, or can be obtained
using the appropriate scs_mfr command-line utilities.

SYNOPSIS
Usage: mqtt_peers.py { -p [-e] | -l [-n HOSTNAME] [-t TOPIC] | -m | -c HOSTNAME TAG SHARED_SECRET TOPIC |
-u HOSTNAME { -s SHARED_SECRET | -t TOPIC } | -r HOSTNAME } [-a] [-i INDENT] [-v]

EXAMPLES
mqtt_peers.py -i scs-bbe-002 scs-be2-2 secret1 scs/production-test/device/alpha-bb-eng-000002/status

csv_reader.py -v mqtt_peers.json | mqtt_peers.py -vie

FILES
~/SCS/aws/mqtt_peers.json

DOCUMENT EXAMPLE
[{"peers": {"scs-bbe-002": {"hostname": "scs-bbe-002", "tag": "scs-be2-2", "shared-secret": "T9o7CMvDB4",
"topic": "south-coast-science-dev/production-test/device/alpha-bb-eng-000002/control"}}]

SEE ALSO
scs_dev/aws_mqtt_control

scs_mfr/aws_project
scs_mfr/shared_secret
scs_mfr/system_id
"""

import json
import sys

from scs_analysis.cmd.cmd_mqtt_peers import CmdMQTTPeers

from scs_core.aws.client.access_key import AccessKey
from scs_core.aws.client.client import Client
from scs_core.aws.manager.s3_manager import S3Manager, S3PersistenceManager

from scs_core.data.json import JSONify

from scs_core.estate.mqtt_device_poller import MQTTDevicePoller
from scs_core.estate.mqtt_peer import MQTTPeer, MQTTPeerSet

from scs_core.sys.logging import Logging

from scs_host.sys.host import Host


# --------------------------------------------------------------------------------------------------------------------

if __name__ == '__main__':

    manager = None
    group = None

    key = None
    document_count = 0

    try:
        # ------------------------------------------------------------------------------------------------------------
        # cmd...

        cmd = CmdMQTTPeers()

        if not cmd.is_valid():
            cmd.print_help(sys.stderr)
            exit(2)

        Logging.config('mqtt_peers', verbose=cmd.verbose)
        logger = Logging.getLogger()

        logger.info(cmd)

        # ------------------------------------------------------------------------------------------------------------
        # resources...

        # PersistenceManager...
        if cmd.aws or cmd.missing:
            if not AccessKey.exists(Host):
                logger.error('AccessKey not available.')
                exit(1)

            try:
                key = AccessKey.load(Host, encryption_key=AccessKey.password_from_user())
            except (KeyError, ValueError):
                logger.error('incorrect password.')
                exit(1)

            client = Client.construct('s3', key)
            resource_client = Client.resource('s3', key)

            if cmd.aws:
                manager = S3PersistenceManager(client, resource_client)
            if cmd.missing:
                manager = S3Manager(client, resource_client)

        else:
            manager = Host

        if not cmd.missing:
            # MQTTPeerSet...
            group = MQTTPeerSet.load(manager)


        # ------------------------------------------------------------------------------------------------------------
        # run...

        # import...
        if cmd.is_import():
            for line in sys.stdin:
                jstr = line.strip()

                if not jstr:
                    continue

                document_count += 1

                peer = MQTTPeer.construct_from_jdict(json.loads(jstr))

                if peer is None:
                    logger.error('invalid document: %s' % jstr)
                    exit(1)

                group.insert(peer)

                if cmd.echo:
                    print(jstr)
                    sys.stdout.flush()

            group.save(manager)
            logger.info('inserted: %d' % document_count)

        # list...
        if cmd.list:
            if group is not None:
                report = group.subset(hostname_substring=cmd.hostname, topic_substring=cmd.topic)

                print(JSONify.dumps(report.peers, indent=cmd.indent))
                logger.info('found: %d' % len(report))

        # create...
        if cmd.is_create():
            peer = MQTTPeer(cmd.create_hostname, cmd.create_tag, cmd.create_shared_secret, cmd.create_topic)
            group.insert(peer)
            group.save(manager)

            print(JSONify.dumps(peer, indent=cmd.indent))

        # update...
        if cmd.is_update():
            report = group.subset(hostname_substring=cmd.update_hostname)

            if len(report) == 0:
                logger.error("no peer matching '%s'." % cmd.update_hostname)
                exit(2)

            if len(report) > 1:
                logger.error("more than one peer matching '%s'." % cmd.update_hostname)
                exit(2)

            peer = report.peer(cmd.update_hostname)

            if cmd.shared_secret:
                peer.shared_secret = cmd.shared_secret

            if cmd.topic:
                peer.topic = cmd.topic

            group.insert(peer)
            group.save(manager)

            print(JSONify.dumps(peer, indent=cmd.indent))

        # delete...
        if cmd.is_remove():
            deleted = group.remove(cmd.remove_hostname)
            document_count = 1 if deleted else 0

            group.save(manager)
            logger.info('deleted: %d' % document_count)

        # missing...
        if cmd.missing:
            reporter = MQTTDevicePoller(Host, manager)
            report = reporter.missing_devices()

            print(JSONify.dumps(report, indent=cmd.indent))
            logger.info('missing: %d' % len(report))

    except KeyboardInterrupt:
        print(file=sys.stderr)




#!/usr/bin/env python3

"""
Created on 28 Jul 2021

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

DESCRIPTION
The baseline utility is use to set the baseline of a device's electrochemical and CO2 sensors according to estimated
low environmental gas concentrations. It does this by downloading device data, finding minimums, then adjusting the
device's baseline offsets by the difference between the estimated and reported values. Any changes in NO2
baselines are incorporated into O3 ('Ox') baselining.

Baselining can be applied to either:

* Val fields - including val.GGG.cnc and val.GGG.vCal (the model inputs)
* Exg fields - of the form exg.val.GGG.cnc or exg.MMM.GGG.cnc (the model outputs)

The device must be online during the baselining operation - a number of MQTT messages are passes between the
utility and the device. If any changes are made to the device's baseline settings, then the gases_sampler process
is restarted. How the restart happens is specified by the --uptake flag.

If the --rehearse flag is used, the baseline utility shows what changes would be made, but does not make the changes.

Operating parameters are specified by the baseline_conf utility, and may be overridden by the baseline utility flags.
Any number of named baseline_conf files may be stored.

The baseline utility requires access_key, aws_api_auth and aws_client_auth to be set.

SYNOPSIS
baseline.py [-a] -c NAME -t DEVICE_TAG -f { V | E } [{ -r | -u COMMAND }] [-s START] [-e END] [-p AGGREGATION]
[-m GAS MINIMUM] [{ -o GAS | -x GAS }] [-v]

EXAMPLES
baseline.py -ac freshfield -t scs-be2-3 -f E -r

FILES
~/SCS/conf/NAME_baseline_conf.json

SEE ALSO
scs_analysis/access_key
scs_analysis/aws_api_auth
scs_analysis/aws_client_auth
scs_analysis/baseline_conf

scs_dev/gases_sampler

scs_mfr/afe_baseline
scs_mfr/gas_baseline
scs_mfr/scd30_baseline
"""

import json

import sys

from scs_analysis.cmd.cmd_baseline import CmdBaseline

from scs_analysis.handler.aws_topic_history_reporter import AWSTopicHistoryReporter

from scs_core.aws.client.access_key import AccessKey
from scs_core.aws.client.api_auth import APIAuth
from scs_core.aws.client.client import Client
from scs_core.aws.client.client_auth import ClientAuth
from scs_core.aws.client.mqtt_client import MQTTClient, MQTTSubscriber

from scs_core.aws.manager.byline_manager import BylineManager
from scs_core.aws.manager.lambda_message_manager import MessageManager
from scs_core.aws.manager.s3_manager import S3PersistenceManager

from scs_core.control.control_handler import ControlHandler

from scs_core.data.datetime import LocalizedDatetime

from scs_core.estate.baseline_conf import BaselineConf
from scs_core.estate.configuration import Configuration
from scs_core.estate.mqtt_peer import MQTTPeerSet

from scs_core.gas.afe_baseline import AFEBaseline
from scs_core.gas.afe_calib import AFECalib
from scs_core.gas.minimum import Minimum

from scs_core.sample.gases_sample import GasesSample

from scs_core.sys.logging import Logging

from scs_host.sys.host import Host


# TODO: optionally specify a closest datetime for low - in order to synchronise multiple devices
# TODO: review Ox handling
# --------------------------------------------------------------------------------------------------------------------

if __name__ == '__main__':

    MQTT_TIMEOUT = 60           # seconds

    logger = None
    key = None
    monitor_auth = None
    mqtt_client = None

    try:
        # ------------------------------------------------------------------------------------------------------------
        # cmd...

        cmd = CmdBaseline()

        if not cmd.is_valid():
            cmd.print_help(sys.stderr)
            exit(2)

        Logging.config('baseline', verbose=cmd.verbose)
        logger = Logging.getLogger()


        # ------------------------------------------------------------------------------------------------------------
        # authentication...

        # AccessKey (for S3PersistenceManager)...
        if not AccessKey.exists(Host):
            logger.error("AccessKey not available.")
            exit(1)

        try:
            key = AccessKey.load(Host, encryption_key=AccessKey.password_from_user())
        except (KeyError, ValueError):
            logger.error("incorrect password")
            exit(1)

        # APIAuth (for BylineManager)...
        api_auth = APIAuth.load(Host)

        if api_auth is None:
            logger.error("APIAuth is not available")
            exit(1)

        # ClientAuth (for MQTTClient)...
        client_auth = ClientAuth.load(Host)

        if client_auth is None:
            logger.error("ClientAuth not available.")
            exit(1)


        # ------------------------------------------------------------------------------------------------------------
        # resources...

        # S3PersistenceManager...
        s3_client = Client.construct('s3', key)
        s3_resource_client = Client.resource('s3', key)

        s3_manager = S3PersistenceManager(s3_client, s3_resource_client)

        # BylineManager...
        byline_manager = BylineManager(api_auth)

        # reporter...
        reporter = AWSTopicHistoryReporter(cmd.verbose)

        # message manager...
        message_manager = MessageManager(api_auth, reporter=reporter)

        # PersistenceManager...
        persistence_manager = s3_manager if cmd.aws else Host


        # ------------------------------------------------------------------------------------------------------------
        # configuration...

        logger.error("configuration...")

        # BaselineConf...
        baseline_conf = BaselineConf.load(persistence_manager, name=cmd.conf_name)

        if baseline_conf is None:
            logger.error("the BaselineConf '%s' is not available." % cmd.conf_name)
            exit(1)

        try:
            baseline_conf = cmd.override(baseline_conf)
        except KeyError as ex:
            logger.error("the gas %s is not in the baseline configuration." % ex)
            exit(2)

        # host name...
        host_tag = Host.name()

        # topics...
        group = byline_manager.find_bylines_for_device(cmd.device_tag)

        if group is None:
            logger.error("no bylines found for %s." % cmd.device_tag)
            exit(1)

        gases_topic = group.latest_topic('/gases')
        control_topic = group.latest_topic('/control')

        if gases_topic is None:
            logger.error("no gases topic found for %s." % cmd.device_tag)
            exit(1)

        if control_topic is None:
            logger.error("no control topic found for %s." % cmd.device_tag)
            exit(1)

        logger.error("gases_topic: %s" % gases_topic)
        logger.error("control_topic: %s" % control_topic)

        # MQTT peer...
        peer_group = MQTTPeerSet.load(s3_manager)
        peer = peer_group.peer_by_tag(cmd.device_tag)

        if peer is None:
            logger.error("no MQTT peer found for tag '%s'." % cmd.device_tag)
            exit(1)

        logger.error("hostname: %s" % peer.hostname)

        # MQTTClient...
        handler = ControlHandler(host_tag, cmd.device_tag)
        subscriber = MQTTSubscriber(control_topic, handler.handle)

        mqtt_client = MQTTClient(subscriber)
        mqtt_client.connect(client_auth, False)

        # Configuration...
        stdout, stderr = handler.publish(mqtt_client, control_topic, ['configuration'], MQTT_TIMEOUT,
                                         peer.shared_secret)

        if stderr:
            logger.error("Configuration cannot be retrieved: %s" % stderr[0])
            exit(1)

        jdict = json.loads(stdout[0])
        device_conf = Configuration.construct_from_jdict(jdict.get('val'))

        ox_index = None if device_conf.afe_id is None else device_conf.afe_id.sensor_index('Ox')

        # AFECAlib...
        if ox_index is None:
            ox_sensor = None

        else:
            stdout, stderr = handler.publish(mqtt_client, control_topic, ['afe_calib'], MQTT_TIMEOUT,
                                             peer.shared_secret)
            if stderr:
                logger.error("AFECAlib cannot be retrieved: %s" % stderr[0])
                exit(1)

            jdict = json.loads(stdout[0])
            afe_calib = AFECalib.construct_from_jdict(jdict)

            afe_baseline = AFEBaseline.null_datum() if device_conf.afe_baseline is None else device_conf.afe_baseline

            ox_baseline = afe_baseline.sensor_baseline(ox_index)
            ox_calib = afe_calib.sensor_calib(ox_index)

            ox_sensor = ox_calib.sensor(ox_baseline)
            # logger.error(ox_sensor)

        logger.error("-")


        # ------------------------------------------------------------------------------------------------------------
        # run...

        # data...
        logger.error("data...")

        now = LocalizedDatetime.now()

        start = baseline_conf.start_datetime(now)
        end = baseline_conf.end_datetime(now)

        logger.error("start: %s end: %s" % (start, end))

        if start == end:
            logger.error("the start and end hours may not be the same.")
            exit(1)

        if end > now:
            logger.error("baselining cannot take place during the test period.")
            exit(1)

        data = list(message_manager.find_for_topic(gases_topic, start, end, None, False, baseline_conf.checkpoint(),
                                                   False, False, False, False, False, None))
        if not data:
            logger.error("no data found for %s." % gases_topic)
            exit(1)

        logger.error("expected: %s retrieved: %s" % (baseline_conf.expected_data_points(start, end), len(data)))
        logger.error("-")

        # corrections...
        logger.error("correction...")

        conf_minimums = baseline_conf.minimums
        correction_applied = False
        no2_correction = None

        for minimum in Minimum.find_minimums(data, cmd.fields):
            if cmd.excludes_gas(minimum.gas):
                continue

            logger.error("-")

            if minimum.gas == cmd.exclude_gas:
                logger.error("%s is excluded - skipping" % minimum.gas)
                continue

            if minimum.gas not in conf_minimums:
                logger.error("%s has no specified minimum - skipping" % minimum.gas)
                continue

            # NO2...
            if minimum.path == 'val.NO2.cnc':
                no2_correction = conf_minimums['NO2'] - minimum.value

            try:
                if minimum.update_already_done(device_conf, end):
                    logger.error("%s has been updated since the latest test period - skipping" % minimum.path)
                    continue

            except ValueError as ex:
                logger.error("sensor with serial number %s is not supported - skipping" % ex)
                continue

            # Ox...
            if minimum.path == 'val.Ox.cnc':
                if no2_correction is None:
                    logger.error('NO2 minimum required for Ox, but none available - skipping')
                    continue

                sample = GasesSample.construct_from_jdict(minimum.sample)

                temp = sample.sht_datum.temp
                no2_sample = sample.electrochem_datum.sns['NO2']
                ox_sample = sample.electrochem_datum.sns['Ox']

                no2_cnc = no2_sample.cnc + no2_correction
                corrected = ox_sensor.datum(temp, ox_sample.we_v, ox_sample.ae_v, no2_cnc=no2_cnc)
                minimum.value = corrected.cnc

            # minimums...
            if minimum.value == conf_minimums[minimum.gas]:
                logger.error("%s matches the specified minimum - skipping" % minimum.path)
                continue

            if minimum.index == 0:
                logger.error("WARNING: the first datum for %s is the minimum value" % minimum.path)

            elif minimum.index == len(data) - 1:
                logger.error("WARNING: the last datum for %s is the minimum value" % minimum.path)

            # logger.error("acting on: %s" % minimum)
            # logger.error("...")

            cmd_tokens = minimum.cmd_tokens(conf_minimums)
            logger.error(' '.join([str(token) for token in cmd_tokens]))

            if cmd.rehearse:
                continue

            stdout, stderr = handler.publish(mqtt_client, control_topic, cmd_tokens, MQTT_TIMEOUT, peer.shared_secret)
            correction_applied = True

            if stderr:
                print(*stderr, sep='\n', file=sys.stderr)
            if stdout:
                print(*stdout, sep='\n', file=sys.stdout)

        if not correction_applied:
            exit(0)

        logger.error("-")

        # reboot...
        logger.error(cmd.uptake)

        handler.publish(mqtt_client, control_topic, [cmd.uptake], MQTT_TIMEOUT, peer.shared_secret)


    # ----------------------------------------------------------------------------------------------------------------
    # end...

    except KeyboardInterrupt:
        print(file=sys.stderr)

    except TimeoutError:
        logger.error("device is not available.")
        exit(1)

    finally:
        if mqtt_client:
            mqtt_client.disconnect()



#!/usr/bin/env python3

"""
Created on 4 Sep 2019

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

source repo: scs_analysis

DESCRIPTION
The sample_nullify utility is used to nullify values where that value - or some other value in the same JSON document -
are outside one or two bounding values. If present, values must be numeric. If not present, the JSON document is passed
to stdout without alteration.

Evaluation follows the rule:

lower bound <= value < upper bound

Both upper and lower bounds are optional. If both are present, then the lower bounding value must be less than the
upper bounding value.

SYNOPSIS
sample_nullify.py -t TARGET_PATH -s SOURCE_PATH [-l LOWER] [-u UPPER] [-v]

EXAMPLES
csv_reader.py -v scs-bgx-405-corrected-2019-04-1min.csv | sample_nullify.py -v -u 80 -s meteo.val.hmd -t proc_PM10 |
csv_writer.py -v scs-bgx-405-corrected-2019-04-1min-r80.csv

DOCUMENT EXAMPLE - INPUT
{"rec": "2019-04-27T02:07:00Z", "opc": {"tag": "scs-bgx-405", "src": "N2",
"val": {"per": 9.9, "pm1": 2.6, "pm2p5": 3.9, "pm10": 7.0}}, "meteo": {"tag": "scs-bgx-405",
"val": {"hmd": 87.6, "tmp": 20.0, "bar": ""}}, "error_model": 4.684,
"proc_PM10": 1.5, "proc_PM2p5": 0.8, "proc_PM1": 0.6}

DOCUMENT EXAMPLE - OUTPUT
{"rec": "2019-04-27T02:07:00Z", "opc": {"tag": "scs-bgx-405", "src": "N2",
"val": {"per": 9.9, "pm1": 2.6, "pm2p5": 3.9, "pm10": 7.0}}, "meteo": {"tag": "scs-bgx-405",
"val": {"hmd": 87.6, "tmp": 20.0, "bar": ""}}, "error_model": 4.684,
"proc_PM10": "", "proc_PM2p5": 0.8, "proc_PM1": 0.6}
"""

import sys

from scs_analysis.cmd.cmd_sample_nullify import CmdSampleNullify

from scs_core.data.json import JSONify
from scs_core.data.path_dict import PathDict


# --------------------------------------------------------------------------------------------------------------------

if __name__ == '__main__':

    document_count = 0
    processed_count = 0
    nullified_count = 0

    # ----------------------------------------------------------------------------------------------------------------
    # cmd...

    cmd = CmdSampleNullify()

    if not cmd.is_valid():
        cmd.print_help(sys.stderr)
        exit(2)

    if cmd.verbose:
        print("sample_nullify: %s" % cmd, file=sys.stderr)
        sys.stderr.flush()

    try:
        # ------------------------------------------------------------------------------------------------------------
        # run...

        for line in sys.stdin:
            jstr = line.strip()
            datum = PathDict.construct_from_jstr(jstr)

            if datum is None:
                continue

            document_count += 1

            paths = datum.paths()
            if cmd.source not in paths or cmd.target not in paths:
                continue

            value_node = datum.node(cmd.source)

            if value_node is not None and value_node != '':
                try:
                    source_value = float(value_node)
                except ValueError:
                    source_value = None
                    print("sample_nullify: invalid numeric value '%s' in %s" % (value_node, jstr), file=sys.stderr)
                    exit(1)

                if (cmd.lower is not None and source_value < cmd.lower) or \
                        (cmd.upper is not None and source_value >= cmd.upper):
                    datum.set_node(cmd.target, None)
                    nullified_count += 1

            print(JSONify.dumps(datum))
            sys.stdout.flush()

            processed_count += 1


    # ----------------------------------------------------------------------------------------------------------------
    # end...

    except KeyError as ex:
        print("sample_nullify: KeyError: %s" % ex, file=sys.stderr)

    except KeyboardInterrupt:
        print(file=sys.stderr)

    finally:
        if cmd.verbose:
            print("sample_nullify: documents: %d processed: %d nullified: %d" %
                  (document_count, processed_count, nullified_count), file=sys.stderr)



"""
Created on 20 Apr 2021

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)
"""

import optparse


# --------------------------------------------------------------------------------------------------------------------

class CmdMonitorAuth(object):
    """unix command line handler"""

    def __init__(self):
        """
        Constructor
        """
        self.__parser = optparse.OptionParser(usage="%prog [{ -s | -d }] [-v]", version="%prog 1.0")

        # commands..
        self.__parser.add_option("--set", "-s", action="store_true", dest="set", default=False,
                                 help="set the authentication interactively")

        self.__parser.add_option("--delete", "-d", action="store_true", dest="delete", default=False,
                                 help="delete the authentication")

        # reporting flag..
        self.__parser.add_option("--verbose", "-v", action="store_true", dest="verbose", default=False,
                                 help="report narrative to stderr")

        self.__opts, self.__args = self.__parser.parse_args()


    # ----------------------------------------------------------------------------------------------------------------

    def is_valid(self):
        if self.set and self.delete:
            return False

        return True

    # ----------------------------------------------------------------------------------------------------------------

    @property
    def set(self):
        return self.__opts.set


    @property
    def delete(self):
        return self.__opts.delete


    @property
    def verbose(self):
        return self.__opts.verbose


    # ----------------------------------------------------------------------------------------------------------------

    def print_help(self, file):
        self.__parser.print_help(file)


    def __str__(self, *args, **kwargs):
        return "CmdMonitorAuth:{set:%s, delete:%s, verbose:%s}" % \
               (self.set, self.delete, self.verbose)



#!/usr/bin/env python3

"""
Created on 2 Apr 2018

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

source repo: scs_analysis

DESCRIPTION
The aws_api_auth utility is used to store or read the endpoint host name and API key required by the South Coast
Science / AWS historic data retrieval system.

SYNOPSIS
aws_api_auth.py [{ [-e ENDPOINT] [-a API_KEY] | -d }] [-v]

EXAMPLES
aws_api_auth.py -e aws.southcoastscience.com -a de92c5ff-b47a-4cc4-a04c-62d684d64a1f

FILES
~/SCS/aws/aws_api_auth.json

DOCUMENT EXAMPLE
{"endpoint": "aws.southcoastscience.com", "api-key": "de92c5ff-b47a-4cc4-a04c-62d684d64a1f"}

SEE ALSO
scs_analysis/aws_topic_history
"""

import sys

from scs_core.aws.client.api_auth import APIAuth
from scs_core.data.json import JSONify

from scs_host.sys.host import Host

from scs_analysis.cmd.cmd_aws_api_auth import CmdAWSAPIAuth


# --------------------------------------------------------------------------------------------------------------------

if __name__ == '__main__':

    # ----------------------------------------------------------------------------------------------------------------
    # cmd...

    cmd = CmdAWSAPIAuth()

    if not cmd.is_valid():
        cmd.print_help(sys.stderr)
        exit(2)

    if cmd.verbose:
        print("aws_api_auth: %s" % cmd, file=sys.stderr)
        sys.stderr.flush()


    # ----------------------------------------------------------------------------------------------------------------
    # resources...

    # APIAuth...
    auth = APIAuth.load(Host)


    # ----------------------------------------------------------------------------------------------------------------
    # run...

    if cmd.set():
        if auth is None and not cmd.is_complete():
            print("aws_api_auth: No configuration is stored - you must therefore set all fields.", file=sys.stderr)
            cmd.print_help(sys.stderr)
            exit(2)

        endpoint = cmd.endpoint if cmd.endpoint else auth.endpoint
        api_key = cmd.api_key if cmd.api_key else auth.api_key

        auth = APIAuth(endpoint, api_key)
        auth.save(Host)

    if cmd.delete:
        auth.delete(Host)
        auth = None

    if auth:
        print(JSONify.dumps(auth))



#!/usr/bin/env python3

"""
Created on 4 Mar 2022

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

source repo: scs_analysis

DESCRIPTION
The sample_regression utility provides a statical report on a single specified field for documents provided in the
input data stream. The specified field normally represents an error or difference betwen two sources.

If the specified field is not present in any of the input documents, the sample_regression utility terminates. If the
field is present but cannot be interpreted as a float, that document is ignored.

The output is a single JSON document. Report fields are:

* Document count
* Minimum
* Mean
* Median
* Maximum
* Variance
* First standard deviation
* Second standard deviation
* Third standard deviation

A minimum of two input documents are required.

SYNOPSIS
sample_stats.py [-p PRECISION] [-v] PATH

EXAMPLES
csv_reader.py -v Mfi_so2_vE_22Q1_results_err.csv | \
sample_stats.py -v -p 3 err.SO2.vE.Urban.22Q1 | \
csv_writer.py -v Mfi_so2_vE_22Q1_results_err_stats.csv

DOCUMENT EXAMPLE - INPUT
{"ref": {"gas": {"SO2": 4.1}}, "SO2": {"vE": {"Urban": {"22Q1": 1.9}}},
"err": {"SO2": {"vE": {"Urban": {"22Q1": -2.2}}}}}
...

DOCUMENT EXAMPLE - OUTPUT
{"err": {"SO2": {"vE": {"Urban": {"22Q1": {"count": 14167, "min": -239.8, "mean": 0.1, "median": 0.1, "max": 145.0,
"var": 19.5, "stdev": 4.4, "stdev2": 8.8, "stdev3": 13.2}}}}}}

RESOURCES
https://en.wikipedia.org/wiki/Standard_deviation
"""

import sys

from scs_analysis.cmd.cmd_sample_stats import CmdSampleStats

from scs_core.data.json import JSONify
from scs_core.data.path_dict import PathDict
from scs_core.data.stats import Stats

from scs_core.sys.logging import Logging


# --------------------------------------------------------------------------------------------------------------------

if __name__ == '__main__':

    document_count = 0
    processed_count = 0

    # ----------------------------------------------------------------------------------------------------------------
    # cmd...

    cmd = CmdSampleStats()

    if not cmd.is_valid():
        cmd.print_help(sys.stderr)
        exit(2)

    Logging.config('sample_stats', verbose=cmd.verbose)
    logger = Logging.getLogger()

    logger.info(cmd)

    try:
        # ------------------------------------------------------------------------------------------------------------
        # run...

        # data...
        values = []

        for line in sys.stdin:
            datum = PathDict.construct_from_jstr(line)

            if datum is None:
                continue

            document_count += 1

            if not datum.has_path(cmd.path):
                logger.error("path '%s' not in datum: %s" % (cmd.path, line.strip()))
                exit(1)

            try:
                values.append(float(datum.node(cmd.path)))
            except ValueError:
                continue

            processed_count += 1

        # validation...
        if len(values) < 2:
            logger.error("at least two valid input documents are required.")
            exit(1)

        # stats...
        stats = Stats.construct(values, prec=cmd.precision)
        logger.info(stats)

        # output...
        stats_dict = PathDict(stats.as_json())
        report = PathDict()

        for stats_path in stats_dict.paths():
            report.append('.'.join((cmd.path, stats_path)), stats_dict.node(stats_path))

        print(JSONify.dumps(report))


    # ----------------------------------------------------------------------------------------------------------------
    # end...

    except KeyboardInterrupt:
        print(file=sys.stderr)

    finally:
        logger.info("documents: %d processed: %d" % (document_count, processed_count))



"""
Created on 9 May 2017

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

source repo: scs_analysis
"""

import optparse


# --------------------------------------------------------------------------------------------------------------------

class CmdMQTTControl(object):
    """unix command line handler"""

    DEFAULT_TIMEOUT = 30                # seconds


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self):
        """
        Constructor
        """
        self.__parser = optparse.OptionParser(usage="%prog { -p HOSTNAME [-a] | -d TAG SHARED_SECRET TOPIC } "
                                                    "{ -i | -r [CMD_TOKENS] } [-t TIMEOUT] [-v]", version="%prog 1.0")

        # compulsory...
        self.__parser.add_option("--peer", "-p", type="string", nargs=1, action="store", dest="peer",
                                 help="use the stored MQTT peer")

        self.__parser.add_option("--device", "-d", type="string", nargs=3, action="store", dest="device",
                                 help="specify a tag, shared secret and topic for the peer")

        # optional...
        self.__parser.add_option("--aws", "-a", action="store_true", dest="aws", default=False,
                                 help="Use AWS S3 instead of local storage for peer")

        self.__parser.add_option("--receipt", "-r", action="store_true", dest="receipt", default=False,
                                 help="wait for receipt from target peer")

        self.__parser.add_option("--interactive", "-i", action="store_true", dest="interactive", default=False,
                                 help="interactive mode (always waits for receipt)")

        self.__parser.add_option("--timeout", "-t", type="int", nargs=1, action="store", dest="timeout",
                                 default=self.DEFAULT_TIMEOUT,
                                 help="receipt timeout in seconds (default %d)" % self.DEFAULT_TIMEOUT)

        self.__parser.add_option("--verbose", "-v", action="store_true", dest="verbose", default=False,
                                 help="report narrative to stderr")

        self.__opts, self.__args = self.__parser.parse_args()


    # ----------------------------------------------------------------------------------------------------------------

    def is_valid(self):
        if self.is_stored_peer() == self.is_device():
            return False

        if self.interactive == self.receipt:
            return False

        if self.interactive and self.cmd_tokens is not None:
            return False

        if self.aws and not self.is_stored_peer():
            return False

        return True


    # ----------------------------------------------------------------------------------------------------------------

    def is_stored_peer(self):
        return self.__opts.peer is not None


    def is_device(self):
        return self.__opts.device is not None


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def peer_hostname(self):
        return self.__opts.peer


    @property
    def aws(self):
        return self.__opts.aws


    @property
    def device_tag(self):
        return None if self.__opts.device is None else self.__opts.device[0]


    @property
    def device_shared_secret(self):
        return None if self.__opts.device is None else self.__opts.device[1]


    @property
    def device_topic(self):
        return None if self.__opts.device is None else self.__opts.device[2]


    @property
    def cmd_tokens(self):
        return self.__args[0].split() if len(self.__args) > 0 else None


    @property
    def receipt(self):
        return self.__opts.receipt


    @property
    def interactive(self):
        return self.__opts.interactive


    @property
    def timeout(self):
        return self.__opts.timeout


    @property
    def verbose(self):
        return self.__opts.verbose


    # ----------------------------------------------------------------------------------------------------------------

    def print_help(self, file):
        self.__parser.print_help(file)


    def __str__(self, *args, **kwargs):
        return "CmdMQTTControl:{peer:%s, device:%s, aws:%s, cmd_tokens:%s, receipt:%s, interactive:%s, " \
               "timeout:%s, verbose:%s}" % \
               (self.__opts.peer, self.__opts.device, self.aws, self.cmd_tokens, self.receipt, self.interactive,
                self.timeout, self.verbose)
